from enum import IntEnum, IntFlag, auto
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Generator,
    NamedTuple,
    Optional,
    Set,
    Tuple,
    Type,
    Union,
    overload,
)

from app_model.types._constants import OperatingSystem

if TYPE_CHECKING:
    from pydantic.annotated_handlers import GetCoreSchemaHandler
    from pydantic_core import core_schema

__all__ = ["KeyCode", "KeyMod", "ScanCode", "KeyChord"]

# TODO:
# https://stackoverflow.com/questions/3202629/where-can-i-find-a-list-of-mac-virtual-key-codes/16125341#16125341

# flake8: noqa
# fmt: off


class KeyCode(IntEnum):
    """Virtual Key Codes, the integer value does not hold any inherent meaning.

    This is the primary internal representation of a key.
    """

    UNKNOWN = 0

    # -----------------------   Writing System Keys   -----------------------
    Backquote = auto()		#	`~ on a US keyboard.
    Backslash = auto()      #	\| on a US keyboard.
    BracketLeft = auto()    #	[{ on a US keyboard.
    BracketRight = auto()   #	]} on a US keyboard.
    Comma = auto()			#	,< on a US keyboard.
    Digit0 = auto()			#	0) on a US keyboard.
    Digit1 = auto()			#	1! on a US keyboard.
    Digit2 = auto()			#	2@ on a US keyboard.
    Digit3 = auto()			#	3# on a US keyboard.
    Digit4 = auto()			#	4$ on a US keyboard.
    Digit5 = auto()			#	5% on a US keyboard.
    Digit6 = auto()			#	6^ on a US keyboard.
    Digit7 = auto()			#	7& on a US keyboard.
    Digit8 = auto()			#	8* on a US keyboard.
    Digit9 = auto()			#	9( on a US keyboard.
    Equal = auto()			#	=+ on a US keyboard.
    IntlBackslash = auto()	#	Located between the left Shift and Z keys. Labelled \| on a UK keyboard.
    KeyA = auto()
    KeyB = auto()
    KeyC = auto()
    KeyD = auto()
    KeyE = auto()
    KeyF = auto()
    KeyG = auto()
    KeyH = auto()
    KeyI = auto()
    KeyJ = auto()
    KeyK = auto()
    KeyL = auto()
    KeyM = auto()
    KeyN = auto()
    KeyO = auto()
    KeyP = auto()
    KeyQ = auto()
    KeyR = auto()
    KeyS = auto()
    KeyT = auto()
    KeyU = auto()
    KeyV = auto()
    KeyW = auto()
    KeyX = auto()
    KeyY = auto()
    KeyZ = auto()
    Minus = auto()			#	-_ on a US keyboard.
    Period = auto()			#	.> on a US keyboard.
    Quote = auto()			#	'" on a US keyboard.
    Semicolon = auto()		#	;: on a US keyboard.
    Slash = auto()			#	/? on a US keyboard.

    # -------------------   Functional Keys   --------------------------------
    Alt = auto()		    # Alt, Option or ⌥.
    Backspace = auto()		# Backspace or ⌫. Labelled Delete on Apple keyboards.
    CapsLock = auto()		# CapsLock or ⇪
    ContextMenu = auto()	# The application context menu key, which is typically found between the right Meta key and the right Control key.
    Ctrl = auto()	        # Control or ⌃
    Enter = auto()		    # Enter or ↵. Labelled Return on Apple keyboards.
    Meta = auto()		    # The Windows, ⌘, Command or other OS symbol key.
    Shift = auto()		    # Shift or ⇧
    Space = auto()		    # (space)
    Tab = auto()		    # Tab or ⇥

    # ----------------------   Control Pad   --------------------------------

    Delete = auto()         # ⌦. The forward delete key. NOT the Delete key on a mac
    End = auto()            # Page Down, End or ↘
    Home = auto()           # Home or ↖
    Insert = auto()         # Insert or Ins. Not present on Apple keyboards.
    PageDown = auto()       # Page Down, PgDn or ⇟
    PageUp = auto()         # Page Up, PgUp or ⇞

    # -----------------------   Arrow Pad   ----------------------------------

    DownArrow = auto()      # ↓
    LeftArrow = auto()      # ←
    RightArrow = auto()     # →
    UpArrow = auto()        # ↑

    # -----------------------   Numpad Section   -----------------------------

    NumLock = auto()            #
    Numpad0 = auto()            # 0
    Numpad1 = auto()            # 1
    Numpad2 = auto()            # 2
    Numpad3 = auto()            # 3
    Numpad4 = auto()            # 4
    Numpad5 = auto()            # 5
    Numpad6 = auto()            # 6
    Numpad7 = auto()            # 7
    Numpad8 = auto()            # 8
    Numpad9 = auto()            # 9
    NumpadAdd = auto()          # +
    NumpadDecimal = auto()      # .
    NumpadDivide = auto()       # /
    NumpadMultiply = auto()     # *
    NumpadSubtract = auto()     # -

    # ---------------------   Function Section   -----------------------------

    Escape = auto() 	        # Esc or ⎋
    F1 = auto()
    F2 = auto()
    F3 = auto()
    F4 = auto()
    F5 = auto()
    F6 = auto()
    F7 = auto()
    F8 = auto()
    F9 = auto()
    F10 = auto()
    F11 = auto()
    F12 = auto()
    PrintScreen = auto()
    ScrollLock = auto()
    PauseBreak = auto()

    def __str__(self) -> str:
        """Get a normalized string representation (constant to all OSes) of this `KeyCode`."""
        return keycode_to_string(self)

    def os_symbol(self, os: Optional[OperatingSystem] = None) -> str:
        """Get a string representation of this `KeyCode` using a symbol/OS specific symbol.

        Some examples:
            * `KeyCode.Enter` is represented by `↵`
            * `KeyCode.Meta` is represented by `⊞` on Windows, `Super` on Linux and `⌘` on MacOS

        If no OS is given, the current detected one is used.
        """
        os = OperatingSystem.current() if os is None else os
        return keycode_to_os_symbol(self, os)

    def os_name(self, os: Optional[OperatingSystem] = None) -> str:
        """Get a string representation of this `KeyCode` using the OS specific naming for the key.

        This differs from `__str__` since with it a normalized representation (constant to all OSes) is given.
        Sometimes these representations coincide but not always! Some examples:
            * `KeyCode.Enter` is represented by `Enter` (`__str__` represents it as `Enter`)
            * `KeyCode.Meta` is represented by `Win` on Windows, `Super` on Linux and `Cmd` on MacOS
            (`__str__` represents it as `Meta`)

        If no OS is given, the current detected one is used.
        """
        os = OperatingSystem.current() if os is None else os
        return keycode_to_os_name(self, os)

    @classmethod
    def from_string(cls, string: str) -> 'KeyCode':
        """Return the `KeyCode` associated with the given string.

        Returns `KeyCode.UNKNOWN` if no `KeyCode` is associated with the string.
        """
        return keycode_from_string(string)

    @classmethod
    def from_event_code(cls, event_code: int) -> 'KeyCode':
        """Return the `KeyCode` associated with the given event code.

        Returns `KeyCode.UNKNOWN` if no `KeyCode` is associated with the event code.
        """
        return _EVENTCODE_TO_KEYCODE.get(event_code, KeyCode.UNKNOWN)


    @classmethod
    def __get_pydantic_core_schema__(
        cls, source: type, handler: 'GetCoreSchemaHandler'
    ) -> 'core_schema.CoreSchema':
        from pydantic_core import core_schema

        return core_schema.no_info_plain_validator_function(cls.validate)

    @classmethod
    def validate(cls, value: Any) -> 'KeyCode':
        if isinstance(value, KeyCode):
            return value
        if isinstance(value, int):
            return cls(value)
        if isinstance(value, str):
            return cls.from_string(value)
        raise TypeError(f'cannot convert type {type(value)!r} to KeyCode')


class ScanCode(IntEnum):
    """Scan codes for the keyboard.

    https://en.wikipedia.org/wiki/Scancode

    These are the scan codes required to conform to the W3C specification for
    KeyboardEvent.code
    
    A scan code is a hardware-specific code that is generated by the keyboard when a key
    is pressed or released. It represents the physical location of a key on the keyboard
    and is unique to each key. A key code, on the other hand, is a higher-level
    representation of a keypress or key release event. They are associated with
    characters, functions, or actions rather than hardware locations.
    As an example, the left and right control keys have the same key code (KeyCode.Ctrl)
    but different scan codes (LeftControl and RightControl).

    https://w3c.github.io/uievents-code/

    commented out lines represent keys that  are optional and may be used by
    implementations to support special keyboards (such as multimedia or
    legacy keyboards).
    """

    UNIDENTIFIED = 0        # This value code should be used when no other value given in this specification is appropriate.

    # -----------------------   Writing System Keys   -----------------------
    # https://w3c.github.io/uievents-code/#key-alphanumeric-writing-system
    # The writing system keys are those that change meaning (i.e., they produce
    # different key values) based on the current locale and keyboard layout.
    # ----------------------------------------------------------------------

    Backquote = auto()		#	`~ on a US keyboard. This is the 半角/全角/漢字 (hankaku/zenkaku/kanji) key on Japanese keyboards
    Backslash = auto()      #	Used for both the US \| (on the 101-key layout) and also for the key located between the " and Enter keys on row C of the 102-, 104- and 106-key layouts. Labelled #~ on a UK (102) keyboard.
    BracketLeft = auto()    #	[{ on a US keyboard.
    BracketRight = auto()   #	]} on a US keyboard.
    Comma = auto()			#	,< on a US keyboard.
    Digit0 = auto()			#	0) on a US keyboard.
    Digit1 = auto()			#	1! on a US keyboard.
    Digit2 = auto()			#	2@ on a US keyboard.
    Digit3 = auto()			#	3# on a US keyboard.
    Digit4 = auto()			#	4$ on a US keyboard.
    Digit5 = auto()			#	5% on a US keyboard.
    Digit6 = auto()			#	6^ on a US keyboard.
    Digit7 = auto()			#	7& on a US keyboard.
    Digit8 = auto()			#	8* on a US keyboard.
    Digit9 = auto()			#	9( on a US keyboard.
    Equal = auto()			#	=+ on a US keyboard.
    IntlBackslash = auto()	#	Located between the left Shift and Z keys. Labelled \| on a UK keyboard.
    IntlRo = auto()			#	Located between the / and right Shift keys. Labelled \ろ (ro) on a Japanese keyboard.
    IntlYen = auto()		#	Located between the = and Backspace keys. Labelled ¥ (yen) on a Japanese keyboard. \/ on a Russian keyboard.
    KeyA = auto()			#	a on a US keyboard. Labelled q on an AZERTY (e.g., French) keyboard.
    KeyB = auto()			#	b on a US keyboard.
    KeyC = auto()			#	c on a US keyboard.
    KeyD = auto()			#	d on a US keyboard.
    KeyE = auto()			#	e on a US keyboard.
    KeyF = auto()			#	f on a US keyboard.
    KeyG = auto()			#	g on a US keyboard.
    KeyH = auto()			#	h on a US keyboard.
    KeyI = auto()			#	i on a US keyboard.
    KeyJ = auto()			#	j on a US keyboard.
    KeyK = auto()			#	k on a US keyboard.
    KeyL = auto()			#	l on a US keyboard.
    KeyM = auto()			#	m on a US keyboard.
    KeyN = auto()			#	n on a US keyboard.
    KeyO = auto()			#	o on a US keyboard.
    KeyP = auto()			#	p on a US keyboard.
    KeyQ = auto()			#	q on a US keyboard. Labelled a on an AZERTY (e.g., French) keyboard.
    KeyR = auto()			#	r on a US keyboard.
    KeyS = auto()			#	s on a US keyboard.
    KeyT = auto()			#	t on a US keyboard.
    KeyU = auto()			#	u on a US keyboard.
    KeyV = auto()			#	v on a US keyboard.
    KeyW = auto()			#	w on a US keyboard. Labelled z on an AZERTY (e.g., French) keyboard.
    KeyX = auto()			#	x on a US keyboard.
    KeyY = auto()			#	y on a US keyboard. Labelled z on a QWERTZ (e.g., German) keyboard.
    KeyZ = auto()			#	z on a US keyboard. Labelled w on an AZERTY (e.g., French) keyboard, and y on a QWERTZ (e.g., German) keyboard.
    Minus = auto()			#	-_ on a US keyboard.
    Period = auto()			#	.> on a US keyboard.
    Quote = auto()			#	'" on a US keyboard.
    Semicolon = auto()		#	;: on a US keyboard.
    Slash = auto()			#	/? on a US keyboard.

    # -------------------   Functional Keys   --------------------------------
    # https://w3c.github.io/uievents-code/#key-alphanumeric-functional
    # The functional keys (not to be confused with the function keys described later)
    # are those keys in the alphanumeric section that provide general editing functions
    # that are common to all locales (like Shift, Tab, Enter and Backspace).
    # With a few exceptions, these keys do not change meaning based on the current
    # keyboard layout.
    # ------------------------------------------------------------------------

    AltLeft = auto()		# Alt, Option or ⌥.
    AltRight = auto()		# Alt, Option or ⌥. This is labelled AltGr key on many keyboard layouts.
    Backspace = auto()		# Backspace or ⌫. Labelled Delete on Apple keyboards.
    CapsLock = auto()		# CapsLock or ⇪
    ContextMenu = auto()	# The application context menu key, which is typically found between the right Meta key and the right Control key.
    ControlLeft = auto()	# Control or ⌃
    ControlRight = auto()	# Control or ⌃
    Enter = auto()		    # Enter or ↵. Labelled Return on Apple keyboards.
    MetaLeft = auto()		# The Windows, ⌘, Command or other OS symbol key.
    MetaRight = auto()		# The Windows, ⌘, Command or other OS symbol key.
    ShiftLeft = auto()		# Shift or ⇧
    ShiftRight = auto()		# Shift or ⇧
    Space = auto()		    # (space)
    Tab = auto()		    # Tab or ⇥
    # Japanese and Korean keyboards.
    Convert = auto()		# Japanese: 変換 (henkan)
    KanaMode = auto()		# Japanese: カタカナ/ひらがな/ローマ字 (katakana/hiragana/romaji)
    NonConvert = auto()		# Japanese: 無変換 (muhenkan)
    # Lang1 = auto()		# Korean: HangulMode 한/영 (han/yeong) Japanese (Mac keyboard): かな (kana)
    # Lang2 = auto()		# Korean: Hanja 한자 (hanja) Japanese (Mac keyboard): 英数 (eisu)
    # Lang3 = auto()		# Japanese (word-processing keyboard): Katakana
    # Lang4 = auto()		# Japanese (word-processing keyboard): Hiragana
    # Lang5 = auto()		# Japanese (word-processing keyboard): Zenkaku/Hankaku

    # ----------------------   Control Pad   --------------------------------
    # https://w3c.github.io/uievents-code/#key-controlpad-section
    # The control pad section of the keyboard is the set of (usually 6) keys that
    # perform navigating and editing operations, for example, Home, PageUp and Insert.
    # ------------------------------------------------------------------------

    Delete = auto()         # ⌦. The forward delete key. Note that on Apple keyboards, the key labelled Delete on the main part of the keyboard should be encoded as "Backspace".
    End = auto()            # Page Down, End or ↘
    Help = auto()           # Help. Not present on standard PC keyboards.
    Home = auto()           # Home or ↖
    Insert = auto()         # Insert or Ins. Not present on Apple keyboards.
    PageDown = auto()       # Page Down, PgDn or ⇟
    PageUp = auto()         # Page Up, PgUp or ⇞

    # -----------------------   Arrow Pad   ----------------------------------
    # https://w3c.github.io/uievents-code/#key-arrowpad-section
    # The arrow pad contains the 4 arrow keys. The keys are commonly arranged in an
    # "upside-down T" configuration.
    # ------------------------------------------------------------------------

    ArrowDown = auto()      # ↓
    ArrowLeft = auto()      # ←
    ArrowRight = auto()     # →
    ArrowUp = auto()        # ↑

    # -----------------------   Numpad Section   -----------------------------
    # https://w3c.github.io/uievents-code/#key-numpad-section
    # The numpad section is the set of keys on the keyboard arranged in a grid like a
    # calculator or mobile phone. This section contains numeric and mathematical
    # operator keys. Laptop computers and compact keyboards will commonly omit
    # these keys to save space.
    # ------------------------------------------------------------------------

    NumLock = auto()                # On the Mac, the "NumLock" code should be used for the numpad Clear key.
    Numpad0 = auto()                # 0 Ins on a keyboard 0 on a phone or remote control
    Numpad1 = auto()                # 1 End on a keyboard 1 or 1 QZ on a phone or remote control
    Numpad2 = auto()                # 2 ↓ on a keyboard 2 ABC on a phone or remote control
    Numpad3 = auto()                # 3 PgDn on a keyboard 3 DEF on a phone or remote control
    Numpad4 = auto()                # 4 ← on a keyboard 4 GHI on a phone or remote control
    Numpad5 = auto()                # 5 on a keyboard 5 JKL on a phone or remote control
    Numpad6 = auto()                # 6 → on a keyboard 6 MNO on a phone or remote control
    Numpad7 = auto()                # 7 Home on a keyboard 7 PQRS or 7 PRS on a phone or remote control
    Numpad8 = auto()                # 8 ↑ on a keyboard 8 TUV on a phone or remote control
    Numpad9 = auto()                # 9 PgUp on a keyboard 9 WXYZ or 9 WXY on a phone or remote control
    NumpadAdd = auto()              # +
    NumpadDecimal = auto()          # . Del. For locales where the decimal separator is "," (e.g., Brazil), this key may generate a ,.
    NumpadDivide = auto()           # /
    NumpadEnter = auto()            #
    NumpadMultiply = auto()         # * on a keyboard. For use with numpads that provide mathematical operations (+, -, * and /). Use  "NumpadStar" for the * key on phones and remote controls.
    NumpadSubtract = auto()         # -
    NumpadEqual = auto()            # =  NOTE: not required to conform to spec.
    # NumpadBackspace = auto()      # Found on the Microsoft Natural Keyboard.
    # NumpadClear = auto()          # C or AC (All Clear). Also for use with numpads that have a Clear key that is separate from the NumLock key. On the Mac, the numpad Clear key should always be encoded as "NumLock".
    # NumpadClearEntry = auto()     # CE (Clear Entry)
    # NumpadComma = auto()          # , (thousands separator). For locales where the thousands separator is a "." (e.g., Brazil), this key may generate a ..
    # NumpadHash = auto()           # # on a phone or remote control device. This key is typically found below the 9 key and to the right of the 0 key.
    # NumpadMemoryAdd = auto()      # M+ Add current entry to the value stored in memory.
    # NumpadMemoryClear = auto()    # MC Clear the value stored in memory.
    # NumpadMemoryRecall = auto()   # MR Replace the current entry with the value stored in memory.
    # NumpadMemoryStore = auto()    # MS Replace the value stored in memory with the current entry.
    # NumpadMemorySubtract = auto() # M- Subtract current entry from the value stored in memory.
    # NumpadParenLeft = auto()      # ( Found on the Microsoft Natural Keyboard.
    # NumpadParenRight = auto()     # ) Found on the Microsoft Natural Keyboard.
    # NumpadStar = auto()           # * on a phone or remote control device. This key is typically found below the 7 key and to the left of the 0 key. Use "NumpadMultiply" for the * key on numeric keypads.

    # ---------------------   Function Section   -----------------------------
    # https://w3c.github.io/uievents-code/#key-function-section
    # The function section runs along the top of the keyboard (above the alphanumeric
    # section) and contains the function keys and a few additional special keys
    # (for example, Esc and Print Screen). A function key is any of the keys labelled
    # F1 ... F12 that an application or operating system can associate with a
    # custom function or action.
    # ------------------------------------------------------------------------

    Escape = auto() 	    # Esc or ⎋
    F1 = auto() 	        # F1
    F2 = auto() 	        # F2
    F3 = auto() 	        # F3
    F4 = auto() 	        # F4
    F5 = auto() 	        # F5
    F6 = auto() 	        # F6
    F7 = auto() 	        # F7
    F8 = auto() 	        # F8
    F9 = auto() 	        # F9
    F10 = auto()            # F10
    F11 = auto()            # F11
    F12 = auto()            # F12
    PrintScreen = auto()    # PrtScr SysRq or Print Screen
    ScrollLock = auto() 	# Scroll Lock
    Pause = auto()          # Pause Break
    # Fn = auto() 	        # Fn This is typically a hardware key that does not generate a separate code. Most keyboards do not place this key in the function section, but it is included here to keep it with related keys.
    # FnLock = auto() 	    # FLock or FnLock. Function Lock key. Found on the Microsoft Natural Keyboard.

    # ---------------------   Media Keys   ----------------------------
    # https://w3c.github.io/uievents-code/#key-media
    # none of these are required to conform to the spec, and are omitted for now

    # ------------ Legacy, Non-Standard and Special Keys --------------
    # https://w3c.github.io/uievents-code/#key-legacy
    # none of these are required to conform to the spec, and are omitted for now

    def __str__(self) -> str:
        return scancode_to_string(self)

    @classmethod
    def from_string(cls, string: str) -> 'ScanCode':
        """Return the KeyCode associated with the given string.

        Returns ScanCode.UNIDENTIFIED if no match is found.
        """
        return scancode_from_string(string)




_EVENTCODE_TO_KEYCODE: Dict[int, KeyCode] = {}
_NATIVE_WINDOWS_VK_TO_KEYCODE: Dict[str, KeyCode] = {}


# build in a closure to prevent modification and declutter namespace
def _build_maps() -> Tuple[
    Callable[[KeyCode], str],
    Callable[[str], KeyCode],
    Callable[[KeyCode, OperatingSystem], str],
    Callable[[KeyCode, OperatingSystem], str],
    Callable[[ScanCode], str],
    Callable[[str], ScanCode],
]:
    class _KM(NamedTuple):
        scancode: ScanCode
        scanstr: str
        keycode: KeyCode
        keystr: str
        eventcode: int
        virtual_key: str

    _ = ''
    _MAPPINGS = [
        _KM(ScanCode.UNIDENTIFIED, 'None', KeyCode.UNKNOWN, 'unknown', 0, 'VK_UNKNOWN'),
        _KM(ScanCode.KeyA, 'KeyA', KeyCode.KeyA, 'A', 65, 'VK_A'),
        _KM(ScanCode.KeyB, 'KeyB', KeyCode.KeyB, 'B', 66, 'VK_B'),
        _KM(ScanCode.KeyC, 'KeyC', KeyCode.KeyC, 'C', 67, 'VK_C'),
        _KM(ScanCode.KeyD, 'KeyD', KeyCode.KeyD, 'D', 68, 'VK_D'),
        _KM(ScanCode.KeyE, 'KeyE', KeyCode.KeyE, 'E', 69, 'VK_E'),
        _KM(ScanCode.KeyF, 'KeyF', KeyCode.KeyF, 'F', 70, 'VK_F'),
        _KM(ScanCode.KeyG, 'KeyG', KeyCode.KeyG, 'G', 71, 'VK_G'),
        _KM(ScanCode.KeyH, 'KeyH', KeyCode.KeyH, 'H', 72, 'VK_H'),
        _KM(ScanCode.KeyI, 'KeyI', KeyCode.KeyI, 'I', 73, 'VK_I'),
        _KM(ScanCode.KeyJ, 'KeyJ', KeyCode.KeyJ, 'J', 74, 'VK_J'),
        _KM(ScanCode.KeyK, 'KeyK', KeyCode.KeyK, 'K', 75, 'VK_K'),
        _KM(ScanCode.KeyL, 'KeyL', KeyCode.KeyL, 'L', 76, 'VK_L'),
        _KM(ScanCode.KeyM, 'KeyM', KeyCode.KeyM, 'M', 77, 'VK_M'),
        _KM(ScanCode.KeyN, 'KeyN', KeyCode.KeyN, 'N', 78, 'VK_N'),
        _KM(ScanCode.KeyO, 'KeyO', KeyCode.KeyO, 'O', 79, 'VK_O'),
        _KM(ScanCode.KeyP, 'KeyP', KeyCode.KeyP, 'P', 80, 'VK_P'),
        _KM(ScanCode.KeyQ, 'KeyQ', KeyCode.KeyQ, 'Q', 81, 'VK_Q'),
        _KM(ScanCode.KeyR, 'KeyR', KeyCode.KeyR, 'R', 82, 'VK_R'),
        _KM(ScanCode.KeyS, 'KeyS', KeyCode.KeyS, 'S', 83, 'VK_S'),
        _KM(ScanCode.KeyT, 'KeyT', KeyCode.KeyT, 'T', 84, 'VK_T'),
        _KM(ScanCode.KeyU, 'KeyU', KeyCode.KeyU, 'U', 85, 'VK_U'),
        _KM(ScanCode.KeyV, 'KeyV', KeyCode.KeyV, 'V', 86, 'VK_V'),
        _KM(ScanCode.KeyW, 'KeyW', KeyCode.KeyW, 'W', 87, 'VK_W'),
        _KM(ScanCode.KeyX, 'KeyX', KeyCode.KeyX, 'X', 88, 'VK_X'),
        _KM(ScanCode.KeyY, 'KeyY', KeyCode.KeyY, 'Y', 89, 'VK_Y'),
        _KM(ScanCode.KeyZ, 'KeyZ', KeyCode.KeyZ, 'Z', 90, 'VK_Z'),
        _KM(ScanCode.Digit1, 'Digit1', KeyCode.Digit1, '1', 49, 'VK_1'),
        _KM(ScanCode.Digit2, 'Digit2', KeyCode.Digit2, '2', 50, 'VK_2'),
        _KM(ScanCode.Digit3, 'Digit3', KeyCode.Digit3, '3', 51, 'VK_3'),
        _KM(ScanCode.Digit4, 'Digit4', KeyCode.Digit4, '4', 52, 'VK_4'),
        _KM(ScanCode.Digit5, 'Digit5', KeyCode.Digit5, '5', 53, 'VK_5'),
        _KM(ScanCode.Digit6, 'Digit6', KeyCode.Digit6, '6', 54, 'VK_6'),
        _KM(ScanCode.Digit7, 'Digit7', KeyCode.Digit7, '7', 55, 'VK_7'),
        _KM(ScanCode.Digit8, 'Digit8', KeyCode.Digit8, '8', 56, 'VK_8'),
        _KM(ScanCode.Digit9, 'Digit9', KeyCode.Digit9, '9', 57, 'VK_9'),
        _KM(ScanCode.Digit0, 'Digit0', KeyCode.Digit0, '0', 48, 'VK_0'),
        _KM(ScanCode.Enter, 'Enter', KeyCode.Enter, 'Enter', 13, 'VK_RETURN'),
        _KM(ScanCode.Escape, 'Escape', KeyCode.Escape, 'Escape', 27, 'VK_ESCAPE'),
        _KM(ScanCode.Backspace, 'Backspace', KeyCode.Backspace, 'Backspace', 8, 'VK_BACK'),
        _KM(ScanCode.Tab, 'Tab', KeyCode.Tab, 'Tab', 9, 'VK_TAB'),
        _KM(ScanCode.Space, 'Space', KeyCode.Space, 'Space', 32, 'VK_SPACE'),
        _KM(ScanCode.Minus, 'Minus', KeyCode.Minus, '-', 189, 'VK_OEM_MINUS'),
        _KM(ScanCode.Equal, 'Equal', KeyCode.Equal, '=', 187, 'VK_OEM_PLUS'),
        _KM(ScanCode.BracketLeft, 'BracketLeft', KeyCode.BracketLeft, '[', 219, 'VK_OEM_4'),
        _KM(ScanCode.BracketRight, 'BracketRight', KeyCode.BracketRight, ']', 221, 'VK_OEM_6'),
        _KM(ScanCode.Backslash, 'Backslash', KeyCode.Backslash, '\\', 220, 'VK_OEM_5'),
        _KM(ScanCode.Semicolon, 'Semicolon', KeyCode.Semicolon, ';', 186, 'VK_OEM_1'),
        _KM(ScanCode.Quote, 'Quote', KeyCode.Quote, "'", 222, 'VK_OEM_7'),
        _KM(ScanCode.Backquote, 'Backquote', KeyCode.Backquote, '`', 192, 'VK_OEM_3'),
        _KM(ScanCode.Comma, 'Comma', KeyCode.Comma, ',', 188, 'VK_OEM_COMMA'),
        _KM(ScanCode.Period, 'Period', KeyCode.Period, '.', 190, 'VK_OEM_PERIOD'),
        _KM(ScanCode.Slash, 'Slash', KeyCode.Slash, '/', 191, 'VK_OEM_2'),
        _KM(ScanCode.CapsLock, 'CapsLock', KeyCode.CapsLock, 'CapsLock', 20, 'VK_CAPITAL'),
        _KM(ScanCode.F1, 'F1', KeyCode.F1, 'F1', 112, 'VK_F1'),
        _KM(ScanCode.F2, 'F2', KeyCode.F2, 'F2', 113, 'VK_F2'),
        _KM(ScanCode.F3, 'F3', KeyCode.F3, 'F3', 114, 'VK_F3'),
        _KM(ScanCode.F4, 'F4', KeyCode.F4, 'F4', 115, 'VK_F4'),
        _KM(ScanCode.F5, 'F5', KeyCode.F5, 'F5', 116, 'VK_F5'),
        _KM(ScanCode.F6, 'F6', KeyCode.F6, 'F6', 117, 'VK_F6'),
        _KM(ScanCode.F7, 'F7', KeyCode.F7, 'F7', 118, 'VK_F7'),
        _KM(ScanCode.F8, 'F8', KeyCode.F8, 'F8', 119, 'VK_F8'),
        _KM(ScanCode.F9, 'F9', KeyCode.F9, 'F9', 120, 'VK_F9'),
        _KM(ScanCode.F10, 'F10', KeyCode.F10, 'F10', 121, 'VK_F10'),
        _KM(ScanCode.F11, 'F11', KeyCode.F11, 'F11', 122, 'VK_F11'),
        _KM(ScanCode.F12, 'F12', KeyCode.F12, 'F12', 123, 'VK_F12'),
        _KM(ScanCode.PrintScreen, 'PrintScreen', KeyCode.PrintScreen, "PrintScreen", 42, "VK_PRINT"),
        _KM(ScanCode.ScrollLock, 'ScrollLock', KeyCode.ScrollLock, 'ScrollLock', 145, 'VK_SCROLL'),
        _KM(ScanCode.Pause, 'Pause', KeyCode.PauseBreak, 'PauseBreak', 19, 'VK_PAUSE'),
        _KM(ScanCode.Insert, 'Insert', KeyCode.Insert, 'Insert', 45, 'VK_INSERT'),
        _KM(ScanCode.Home, 'Home', KeyCode.Home, 'Home', 36, 'VK_HOME'),
        _KM(ScanCode.PageUp, 'PageUp', KeyCode.PageUp, 'PageUp', 33, 'VK_PRIOR'),
        _KM(ScanCode.Delete, 'Delete', KeyCode.Delete, 'Delete', 46, 'VK_DELETE'),
        _KM(ScanCode.End, 'End', KeyCode.End, 'End', 35, 'VK_END'),
        _KM(ScanCode.PageDown, 'PageDown', KeyCode.PageDown, 'PageDown', 34, 'VK_NEXT'),
        _KM(ScanCode.ArrowRight, 'ArrowRight', KeyCode.RightArrow, 'Right', 39, 'VK_RIGHT'),
        _KM(ScanCode.ArrowLeft, 'ArrowLeft', KeyCode.LeftArrow, 'Left', 37, 'VK_LEFT'),
        _KM(ScanCode.ArrowDown, 'ArrowDown', KeyCode.DownArrow, 'Down', 40, 'VK_DOWN'),
        _KM(ScanCode.ArrowUp, 'ArrowUp', KeyCode.UpArrow, 'Up', 38, 'VK_UP'),
        _KM(ScanCode.NumLock, 'NumLock', KeyCode.NumLock, 'NumLock', 144, 'VK_NUMLOCK'),
        _KM(ScanCode.NumpadDivide, 'NumpadDivide', KeyCode.NumpadDivide, 'NumPad_Divide', 111, 'VK_DIVIDE'),
        _KM(ScanCode.NumpadMultiply, 'NumpadMultiply', KeyCode.NumpadMultiply, 'NumPad_Multiply', 106, 'VK_MULTIPLY'),
        _KM(ScanCode.NumpadSubtract, 'NumpadSubtract', KeyCode.NumpadSubtract, 'NumPad_Subtract', 109, 'VK_SUBTRACT'),
        _KM(ScanCode.NumpadAdd, 'NumpadAdd', KeyCode.NumpadAdd, 'NumPad_Add', 107, 'VK_ADD'),
        _KM(ScanCode.NumpadEnter, 'NumpadEnter', KeyCode.Enter, _, 0, _),
        _KM(ScanCode.Numpad1, 'Numpad1', KeyCode.Numpad1, 'NumPad1', 97, 'VK_NUMPAD1'),
        _KM(ScanCode.Numpad2, 'Numpad2', KeyCode.Numpad2, 'NumPad2', 98, 'VK_NUMPAD2'),
        _KM(ScanCode.Numpad3, 'Numpad3', KeyCode.Numpad3, 'NumPad3', 99, 'VK_NUMPAD3'),
        _KM(ScanCode.Numpad4, 'Numpad4', KeyCode.Numpad4, 'NumPad4', 100, 'VK_NUMPAD4'),
        _KM(ScanCode.Numpad5, 'Numpad5', KeyCode.Numpad5, 'NumPad5', 101, 'VK_NUMPAD5'),
        _KM(ScanCode.Numpad6, 'Numpad6', KeyCode.Numpad6, 'NumPad6', 102, 'VK_NUMPAD6'),
        _KM(ScanCode.Numpad7, 'Numpad7', KeyCode.Numpad7, 'NumPad7', 103, 'VK_NUMPAD7'),
        _KM(ScanCode.Numpad8, 'Numpad8', KeyCode.Numpad8, 'NumPad8', 104, 'VK_NUMPAD8'),
        _KM(ScanCode.Numpad9, 'Numpad9', KeyCode.Numpad9, 'NumPad9', 105, 'VK_NUMPAD9'),
        _KM(ScanCode.Numpad0, 'Numpad0', KeyCode.Numpad0, 'NumPad0', 96, 'VK_NUMPAD0'),
        _KM(ScanCode.NumpadDecimal, 'NumpadDecimal', KeyCode.NumpadDecimal, 'NumPad_Decimal', 110, 'VK_DECIMAL'),
        _KM(ScanCode.IntlBackslash, 'IntlBackslash', KeyCode.IntlBackslash, 'OEM_102', 226, 'VK_OEM_102'),
        _KM(ScanCode.ContextMenu, 'ContextMenu', KeyCode.ContextMenu, 'ContextMenu', 93, _),
        _KM(ScanCode.NumpadEqual, 'NumpadEqual', KeyCode.UNKNOWN, _, 0, _),
        _KM(ScanCode.Help, 'Help', KeyCode.UNKNOWN, _, 0, _),
        _KM(ScanCode.IntlRo, 'IntlRo', KeyCode.UNKNOWN, _, 193, 'VK_ABNT_C1'),
        _KM(ScanCode.KanaMode, 'KanaMode', KeyCode.UNKNOWN, _, 0, _),
        _KM(ScanCode.IntlYen, 'IntlYen', KeyCode.UNKNOWN, _, 0, _),
        _KM(ScanCode.Convert, 'Convert', KeyCode.UNKNOWN, _, 0, _),
        _KM(ScanCode.NonConvert, 'NonConvert', KeyCode.UNKNOWN, _, 0, _),
        _KM(ScanCode.UNIDENTIFIED, _, KeyCode.Ctrl, 'Ctrl', 17, 'VK_CONTROL'),
        _KM(ScanCode.UNIDENTIFIED, _, KeyCode.Shift, 'Shift', 16, 'VK_SHIFT'),
        _KM(ScanCode.UNIDENTIFIED, _, KeyCode.Alt, 'Alt', 18, 'VK_MENU'),
        _KM(ScanCode.UNIDENTIFIED, _, KeyCode.Meta, 'Meta', 0, 'VK_COMMAND'),
        _KM(ScanCode.ControlLeft, 'ControlLeft', KeyCode.Ctrl, _, 0, 'VK_LCONTROL'),
        _KM(ScanCode.ShiftLeft, 'ShiftLeft', KeyCode.Shift, _, 0, 'VK_LSHIFT'),
        _KM(ScanCode.AltLeft, 'AltLeft', KeyCode.Alt, _, 0, 'VK_LMENU'),
        _KM(ScanCode.MetaLeft, 'MetaLeft', KeyCode.Meta, _, 0, 'VK_LWIN'),
        _KM(ScanCode.ControlRight, 'ControlRight', KeyCode.Ctrl, _, 0, 'VK_RCONTROL'),
        _KM(ScanCode.ShiftRight, 'ShiftRight', KeyCode.Shift, _, 0, 'VK_RSHIFT'),
        _KM(ScanCode.AltRight, 'AltRight', KeyCode.Alt, _, 0, 'VK_RMENU'),
        _KM(ScanCode.MetaRight, 'MetaRight', KeyCode.Meta, _, 0, 'VK_RWIN'),
    ]

    SCANCODE_TO_STRING: Dict[ScanCode, str] = {}
    SCANCODE_FROM_LOWERCASE_STRING: Dict[str, ScanCode] = {}

    KEYCODE_TO_STRING: Dict[KeyCode, str] = {}
    KEYCODE_FROM_LOWERCASE_STRING: Dict[str, KeyCode] = {
        # two special cases for assigning os-specific strings to the meta key
        'win': KeyCode.Meta,
        'cmd': KeyCode.Meta,
    }

    # key symbols on all platforms
    KEY_SYMBOLS: dict[KeyCode, str] = {
        KeyCode.Shift: "⇧",
        KeyCode.LeftArrow: "←",
        KeyCode.RightArrow: "→",
        KeyCode.UpArrow: "↑",
        KeyCode.DownArrow: "↓",
        KeyCode.Backspace: "⌫",
        KeyCode.Delete: "⌦",
        KeyCode.Tab: "⇥",
        KeyCode.Escape: "⎋",
        KeyCode.Enter: "↵",
        KeyCode.Space: "␣",
        KeyCode.CapsLock: "⇪",
    }
    # key symbols mappings per platform
    OS_KEY_SYMBOLS: dict[OperatingSystem, dict[KeyCode, str]] = {
        OperatingSystem.WINDOWS: {**KEY_SYMBOLS, KeyCode.Meta: "⊞"},
        OperatingSystem.LINUX: {**KEY_SYMBOLS, KeyCode.Meta: "Super"},
        OperatingSystem.MACOS: {
            **KEY_SYMBOLS,
            KeyCode.Ctrl: "⌃",
            KeyCode.Alt: "⌥",
            KeyCode.Meta: "⌘",
        },
    }

    # key names mappings per platform
    OS_KEY_NAMES: dict[OperatingSystem, dict[KeyCode, str]] = {
        OperatingSystem.WINDOWS: {KeyCode.Meta: "Win"},
        OperatingSystem.LINUX: {KeyCode.Meta: "Super"},
        OperatingSystem.MACOS: {
            KeyCode.Ctrl: "Control",
            KeyCode.Alt: "Option",
            KeyCode.Meta: "Cmd",
        },
    }

    seen_scancodes: Set[ScanCode] = set()
    seen_keycodes: Set[KeyCode] = set()
    for i, km in enumerate(_MAPPINGS):
        if km.scancode not in seen_scancodes:
            seen_scancodes.add(km.scancode)
            SCANCODE_TO_STRING[km.scancode] = km.scanstr
            SCANCODE_FROM_LOWERCASE_STRING[km.scanstr.lower()] = km.scancode
        if km.keycode not in seen_keycodes:
            seen_keycodes.add(km.keycode)
            if not km.keystr:  # pragma: no cover
                raise ValueError(
                    f"String representation missing for key code {km.keycode!r} "
                    f"around scan code {km.scancode!r} at line {i + 1}"
                )
            KEYCODE_TO_STRING[km.keycode] = km.keystr
            KEYCODE_FROM_LOWERCASE_STRING[km.keystr.lower()] = km.keycode
        if km.eventcode:
            _EVENTCODE_TO_KEYCODE[km.eventcode] = km.keycode
        if km.virtual_key:
            _NATIVE_WINDOWS_VK_TO_KEYCODE[km.virtual_key] = km.keycode

    def _keycode_to_string(keycode: KeyCode) -> str:
        """Return the string representation of a KeyCode."""
        # sourcery skip
        return KEYCODE_TO_STRING.get(keycode, "")

    def _keycode_from_string(keystr: str) -> KeyCode:
        """Return KeyCode for a given string."""
        # sourcery skip
        return KEYCODE_FROM_LOWERCASE_STRING.get(str(keystr).lower(), KeyCode.UNKNOWN)

    def _keycode_to_os_symbol(keycode: KeyCode, os: OperatingSystem) -> str:
        """Return key symbol for an OS for a given KeyCode."""
        if keycode in (symbols := OS_KEY_SYMBOLS.get(os, {})):
            return symbols[keycode]
        return str(keycode)

    def _keycode_to_os_name(keycode: KeyCode, os: OperatingSystem) -> str:
        """Return key name for an OS for a given KeyCode."""
        if keycode in (names := OS_KEY_NAMES.get(os, {})):
            return names[keycode]
        return str(keycode)

    def _scancode_to_string(scancode: ScanCode) -> str:
        """Return the string representation of a ScanCode."""
        # sourcery skip
        return SCANCODE_TO_STRING.get(scancode, "")

    def _scancode_from_string(scanstr: str) -> ScanCode:
        """Return ScanCode for a given string."""
        # sourcery skip
        return SCANCODE_FROM_LOWERCASE_STRING.get(
            str(scanstr).lower(), ScanCode.UNIDENTIFIED
        )

    return (
        _keycode_to_string,
        _keycode_from_string,
        _keycode_to_os_symbol,
        _keycode_to_os_name,
        _scancode_to_string,
        _scancode_from_string,
    )


(
    keycode_to_string,
    keycode_from_string,
    keycode_to_os_symbol,
    keycode_to_os_name,
    scancode_to_string,
    scancode_from_string,
) = _build_maps()


# fmt: on

# Keys with modifiers are expressed
# with a 16-bit binary encoding
#
#    1111 11
#    5432 1098 7654 3210
#    ---- CSAW KKKK KKKK
#  C = bit 11   -> ctrlCmd flag
#  S = bit 10   -> shift flag
#  A = bit 9    -> alt flag
#  W = bit 8    -> winCtrl flag
#  K = bits 0-7 -> key code


class KeyMod(IntFlag):
    """A Flag indicating keyboard modifiers."""

    NONE = 0
    CtrlCmd = 1 << 11  # command on a mac, control on windows
    Shift = 1 << 10  # shift key
    Alt = 1 << 9  # alt option
    WinCtrl = 1 << 8  # meta key on windows, ctrl key on mac

    @overload  # type: ignore
    def __or__(self, other: "KeyMod") -> "KeyMod": ...

    @overload
    def __or__(self, other: KeyCode) -> "KeyCombo": ...

    @overload
    def __or__(self, other: int) -> int: ...

    def __or__(  # pyright: ignore[reportIncompatibleMethodOverride]
        self, other: Union["KeyMod", KeyCode, int]
    ) -> Union["KeyMod", "KeyCombo", int]:
        if isinstance(other, self.__class__):
            return self.__class__(self._value_ | other._value_)
        if isinstance(other, KeyCode):
            return KeyCombo(self, other)
        return NotImplemented  # pragma: no cover


class KeyCombo(int):
    """KeyCombo is an integer combination of one or more.

    [`KeyMod`][app_model.types.KeyMod] and [`KeyCode`][app_model.types.KeyCode].
    """

    def __new__(
        cls: Type["KeyCombo"], modifiers: KeyMod, key: KeyCode = KeyCode.UNKNOWN
    ) -> "KeyCombo":
        return super().__new__(cls, int(modifiers) | int(key))

    def __init__(self, modifiers: KeyMod, key: KeyCode = KeyCode.UNKNOWN):
        self._modifiers = modifiers
        self._key = key

    def __repr__(self) -> str:
        name = self.__class__.__name__
        mods_repr = repr(self._modifiers).split(":", 1)[0].split(".", 1)[1]
        return f"<{name}.{mods_repr}|{self._key.name}: {int(self)}>"


class KeyChord(int):
    """KeyChord is an integer combination of two key combos.

    It could be two [`KeyCombo`][app_model.types.KeyCombo]
    [`KeyCode`][app_model.types.KeyCode], or [int][].

    Parameters
    ----------
    first_part : KeyCombo | int
        The first part of the chord.
    second_part : KeyCombo | int
        The second part of the chord.
    """

    def __new__(cls: Type["KeyChord"], first_part: int, second_part: int) -> "KeyChord":
        # shift the second part 16 bits to the left
        chord_part = (second_part & 0x0000FFFF) << 16
        # then combine then to make the full chord
        return super().__new__(cls, first_part | chord_part)

    def __init__(self, first_part: int, second_part: int):
        self._first_part = first_part
        self._second_part = second_part

    def __repr__(self) -> str:
        return f"KeyChord({self._first_part!r}, {self._second_part!r})"



from typing import Any, Callable, Optional, TypedDict, TypeVar, Union

from pydantic import Field, model_validator

from app_model import expressions

from ._base import _BaseModel
from ._constants import KeyBindingSource, OperatingSystem
from ._keys import StandardKeyBinding

KeyEncoding = Union[int, str]
M = TypeVar("M")

_OS = OperatingSystem.current()
_WIN = _OS.is_windows
_MAC = _OS.is_mac
_LINUX = _OS.is_linux


class KeyBindingRule(_BaseModel):
    """Data representing a keybinding and when it should be active.

    This model lacks a corresponding command. That gets linked up elsewhere,
    such as below in `Action`.

    Values can be expressed as either a string (e.g. `"Ctrl+O"`) or an integer, using
    combinations of [`KeyMod`][app_model.types.KeyMod] and
    [`KeyCode`][app_model.types.KeyCode], (e.g. `KeyMod.CtrlCmd | KeyCode.KeyO`).
    """

    primary: Optional[KeyEncoding] = Field(
        default=None, description="(Optional) Key combo, (e.g. Ctrl+O)."
    )
    win: Optional[KeyEncoding] = Field(
        default=None, description="(Optional) Windows specific key combo."
    )
    mac: Optional[KeyEncoding] = Field(
        default=None, description="(Optional) MacOS specific key combo."
    )
    linux: Optional[KeyEncoding] = Field(
        default=None, description="(Optional) Linux specific key combo."
    )
    when: Optional[expressions.Expr] = Field(
        default=None,
        description="(Optional) Condition when the keybingding is active.",
    )
    weight: int = Field(
        default=0,
        description="Internal weight used to sort keybindings. "
        "This is not part of the plugin schema",
    )
    source: KeyBindingSource = Field(
        default=KeyBindingSource.APP,
        description="Who registered the keybinding. Used to sort keybindings.",
    )

    def _bind_to_current_platform(self) -> Optional[KeyEncoding]:
        if _WIN and self.win:
            return self.win
        if _MAC and self.mac:
            return self.mac
        if _LINUX and self.linux:
            return self.linux
        return self.primary

    @model_validator(mode="wrap")
    @classmethod
    def _model_val(
        cls: type[M], v: Any, handler: Callable[[Any], M]
    ) -> "KeyBindingRule":
        if isinstance(v, StandardKeyBinding):
            return v.to_keybinding_rule()
        return handler(v)  # type: ignore


class KeyBindingRuleDict(TypedDict, total=False):
    """Typed dict for KeyBindingRule kwargs."""

    primary: Optional[str]
    win: Optional[str]
    linux: Optional[str]
    mac: Optional[str]
    weight: int
    when: Optional[expressions.Expr]


KeyBindingRuleOrDict = Union[KeyBindingRule, KeyBindingRuleDict]



from typing import (
    Any,
    Optional,
    TypedDict,
    Union,
)

from pydantic import Field, field_validator, model_validator

from app_model import expressions

from ._base import _BaseModel
from ._command_rule import CommandRule
from ._icon import Icon


class MenuItemBase(_BaseModel):
    """Data representing where and when a menu item should be shown."""

    when: Optional[expressions.Expr] = Field(
        default=None,
        description="(Optional) Condition which must be true to show the item.",
    )
    group: Optional[str] = Field(
        default=None,
        description="(Optional) Menu group to which this item should be added. Menu "
        "groups are sortable strings (like `'1_cutandpaste'`). 'navigation' is a "
        "special group that always appears at the top of a menu.  If not provided, "
        "the item is added in the last group of the menu.",
    )
    order: Optional[float] = Field(
        default=None,
        description="(Optional) Order of the item *within* its group. Note, order is "
        "not part of the plugin schema, plugins may provide it using the group key "
        "and the syntax 'group@order'.  If not provided, items are sorted by title.",
    )

    @classmethod
    def _validate(cls: type["MenuItemBase"], v: Any) -> "MenuItemBase":
        """Validate icon."""
        if isinstance(v, MenuItemBase):
            return v
        if isinstance(v, dict):
            if "command" in v:
                return MenuItem(**v)
            if "id" in v:
                return MenuRule(**v)
            if "submenu" in v:
                return SubmenuItem(**v)
        raise ValueError(f"Invalid menu item: {v!r}", cls)  # pragma: no cover


class MenuRule(MenuItemBase):
    """A MenuRule defines a menu location and conditions for presentation.

    It does not define an actual command. That is done in either `MenuItem` or `Action`.
    """

    id: str = Field(..., description="Menu in which to place this item.")

    @model_validator(mode="before")
    @classmethod
    def _validate_model(cls, v: Any) -> Any:
        """If a single string is provided, convert to a dict with `id` key."""
        return {"id": v} if isinstance(v, str) else v


class MenuItem(MenuItemBase):
    """Combination of a Command and conditions for menu presentation.

    This object is mostly constructed by `register_action` right before menu item
    registration.
    """

    command: CommandRule = Field(
        ...,
        description="CommandRule to execute when this menu item is selected.",
    )
    alt: Optional[CommandRule] = Field(
        default=None,
        description="(Optional) Alternate command to execute when this menu item is "
        "selected, (e.g. when the Alt-key is held when opening the menu)",
    )

    @field_validator("command")
    def _simplify_command_rule(cls, v: Any) -> CommandRule:
        if isinstance(v, CommandRule):
            return v._as_command_rule()
        raise TypeError("command must be a CommandRule")  # pragma: no cover


class SubmenuItem(MenuItemBase):
    """Point to another Menu that will be displayed as a submenu."""

    submenu: str = Field(..., description="Menu to insert as a submenu.")
    title: str = Field(..., description="Title of this submenu, shown in the UI.")
    icon: Optional[Icon] = Field(
        default=None,
        description="(Optional) Icon used to represent this submenu. "
        "These may be [iconify keys](https://icon-sets.iconify.design), "
        "such as `fa6-solid:arrow-down`, or "
        "[superqt.fonticon](https://pyapp-kit.github.io/superqt/utilities/fonticon/)"
        " keys, such as `fa6s.arrow_down`",
    )
    enablement: Optional[expressions.Expr] = Field(
        default=None,
        description="(Optional) Condition which must be true to enable the submenu. "
        "Disabled submenus appear grayed out in the UI, and cannot be selected. By "
        "default, submenus are enabled.",
    )


class MenuRuleDict(TypedDict, total=False):
    """Typed dict for MenuRule kwargs.

    This mimics the pydantic `MenuRule` interface, but allows you to pass in a dict
    """

    when: Optional[expressions.Expr]
    group: str
    order: Optional[float]
    id: str


MenuRuleOrDict = Union[MenuRule, MenuRuleDict]
MenuOrSubmenu = Union[MenuItem, SubmenuItem]



from __future__ import annotations

import contextlib
from collections.abc import Collection, Iterable, Mapping, Sequence
from typing import TYPE_CHECKING, cast

from qtpy.QtWidgets import QApplication, QMenu, QMenuBar, QToolBar

from app_model import Application
from app_model.types import SubmenuItem

from ._qaction import QCommandRuleAction, QMenuItemAction
from ._util import to_qicon

try:
    from qtpy import QT6
except ImportError:
    QT6 = False

if TYPE_CHECKING:
    from qtpy.QtWidgets import QAction, QWidget


class QModelMenu(QMenu):
    """QMenu for a menu_id in an `app_model` MenusRegistry.

    Parameters
    ----------
    menu_id : str
        Menu ID to look up in the registry.
    app : Application | str
        Application instance or name of application instance.
    title : str | None
        Optional title for the menu, by default None
    parent : QWidget | None
        Optional parent widget, by default None
    """

    def __init__(
        self,
        menu_id: str,
        app: Application | str,
        title: str | None = None,
        parent: QWidget | None = None,
    ) -> None:
        QMenu.__init__(self, parent)

        # NOTE: code duplication with QModelToolBar, but Qt mixins and multiple
        # inheritance are problematic for some versions of Qt, and for typing
        assert isinstance(menu_id, str), f"Expected str, got {type(menu_id)!r}"
        self._menu_id = menu_id
        self._app = Application.get_or_create(app) if isinstance(app, str) else app
        self.setObjectName(menu_id)
        self.rebuild()
        self._app.menus.menus_changed.connect(self._on_registry_changed)
        self.destroyed.connect(self._disconnect)
        # ----------------------

        if title is not None:
            self.setTitle(title)
        self.aboutToShow.connect(self._on_about_to_show)

    def findAction(self, object_name: str) -> QAction | QModelMenu | None:
        """Find an action by its ObjectName.

        Parameters
        ----------
        object_name : str
            Action ID to find. Note that `QCommandAction` have `ObjectName` set
            to their `command.id`
        """
        return _find_action(self.actions(), object_name)

    def update_from_context(self, ctx: Mapping[str, object]) -> None:
        """Update the enabled/visible state of each menu item with `ctx`.

        See `app_model.expressions` for details on expressions.

        Parameters
        ----------
        ctx : Mapping
            A namespace that will be used to `eval()` the `'enablement'` and
            `'when'` expressions provided for each action in the menu.
            *ALL variables used in these expressions must either be present in
            the `ctx` dict, or be builtins*.
        """
        _update_from_context(self.actions(), ctx)

    def rebuild(
        self, include_submenus: bool = True, exclude: Collection[str] | None = None
    ) -> None:
        """Rebuild menu by looking up self._menu_id in menu_registry."""
        _rebuild(
            menu=self,
            app=self._app,
            menu_id=self._menu_id,
            include_submenus=include_submenus,
            exclude=exclude,
        )

    def _on_about_to_show(self) -> None:
        for action in self.actions():
            if isinstance(action, QCommandRuleAction):
                action._refresh()

    def _disconnect(self) -> None:
        self._app.menus.menus_changed.disconnect(self._on_registry_changed)

    def _on_registry_changed(self, changed_ids: set[str]) -> None:
        if self._menu_id in changed_ids:
            # if this (sub)menu has been removed from the registry,
            # we may hit a RuntimeError when trying to rebuild it.
            with contextlib.suppress(RuntimeError):
                self.rebuild()


class QModelSubmenu(QModelMenu):
    """QMenu for a menu_id in an `app_model` MenusRegistry.

    Parameters
    ----------
    submenu : SubmenuItem
        SubmenuItem for which to create a QMenu.
    app : Application | str
        Application instance or name of application instance.
    parent : QWidget | None
        Optional parent widget, by default None
    """

    def __init__(
        self,
        submenu: SubmenuItem,
        app: Application | str,
        parent: QWidget | None = None,
    ) -> None:
        assert isinstance(submenu, SubmenuItem), f"Expected str, got {type(submenu)!r}"
        self._submenu = submenu
        super().__init__(
            menu_id=submenu.submenu, app=app, title=submenu.title, parent=parent
        )
        if submenu.icon:
            self.setIcon(to_qicon(submenu.icon))

    def update_from_context(self, ctx: Mapping[str, object]) -> None:
        """Update the enabled state of this menu item from `ctx`."""
        super().update_from_context(ctx)
        self.setEnabled(expr.eval(ctx) if (expr := self._submenu.enablement) else True)
        # TODO: ... visibility needs to be controlled at the level of placement
        # in the submenu.  consider only using the `when` expression
        # self.setVisible(expr.eval(ctx) if (expr := self._submenu.when) else True)


class QModelToolBar(QToolBar):
    """QToolBar that is built from a list of model menu ids.

    Parameters
    ----------
    menu_id : str
        Menu ID to look up in the registry.
    app : Application | str
        Application instance or name of application instance.
    exclude : Collection[str] | None
        Optional list of menu ids to exclude from the toolbar, by default None
    title : str | None
        Optional title for the menu, by default None
    parent : QWidget | None
        Optional parent widget, by default None
    """

    def __init__(
        self,
        menu_id: str,
        app: Application | str,
        *,
        exclude: Collection[str] | None = None,
        title: str | None = None,
        parent: QWidget | None = None,
    ) -> None:
        self._exclude = exclude
        QToolBar.__init__(self, parent)

        # NOTE: code duplication with QModelMenu, but Qt mixins and multiple
        # inheritance are problematic for some versions of Qt, and for typing
        assert isinstance(menu_id, str), f"Expected str, got {type(menu_id)!r}"
        self._menu_id = menu_id
        self._app = Application.get_or_create(app) if isinstance(app, str) else app
        self.setObjectName(menu_id)
        self.rebuild()
        self._app.menus.menus_changed.connect(self._on_registry_changed)
        self.destroyed.connect(self._disconnect)
        # ----------------------

        if title is not None:
            self.setWindowTitle(title)

    def addMenu(self, menu: QMenu) -> None:
        """No-op for toolbar."""

    def findAction(self, object_name: str) -> QAction | QModelMenu | None:
        """Find an action by its ObjectName.

        Parameters
        ----------
        object_name : str
            Action ID to find. Note that `QCommandAction` have `ObjectName` set
            to their `command.id`
        """
        return _find_action(self.actions(), object_name)

    def update_from_context(self, ctx: Mapping[str, object]) -> None:
        """Update the enabled/visible state of each menu item with `ctx`.

        See `app_model.expressions` for details on expressions.

        Parameters
        ----------
        ctx : Mapping
            A namespace that will be used to `eval()` the `'enablement'` and
            `'when'` expressions provided for each action in the menu.
            *ALL variables used in these expressions must either be present in
            the `ctx` dict, or be builtins*.
        """
        _update_from_context(self.actions(), ctx)

    def rebuild(
        self, include_submenus: bool = True, exclude: Collection[str] | None = None
    ) -> None:
        """Rebuild toolbar by looking up self._menu_id in menu_registry."""
        _rebuild(
            menu=self,
            app=self._app,
            menu_id=self._menu_id,
            include_submenus=include_submenus,
            exclude=self._exclude if exclude is None else exclude,
        )

    def _disconnect(self) -> None:
        self._app.menus.menus_changed.disconnect(self._on_registry_changed)

    def _on_registry_changed(self, changed_ids: set[str]) -> None:
        if self._menu_id in changed_ids:
            self.rebuild()


class QModelMenuBar(QMenuBar):
    """QMenuBar that is built from a list of model menu ids.

    Parameters
    ----------
    menus : Mapping[str, str] | Sequence[str | tuple[str, str]]
        A mapping of menu ids to menu titles or a sequence of menu ids.
    app : Application | str
        Application instance or name of application instance.
    parent : QWidget | None
        Optional parent widget, by default None
    """

    def __init__(
        self,
        menus: Mapping[str, str] | Sequence[str | tuple[str, str]],
        app: Application | str,
        parent: QWidget | None = None,
    ) -> None:
        super().__init__(parent)

        menu_items = menus.items() if isinstance(menus, Mapping) else menus
        for item in menu_items:
            id_, title = item if isinstance(item, tuple) else (item, item.title())
            self.addMenu(QModelMenu(id_, app, title, self))

    def update_from_context(self, ctx: Mapping[str, object]) -> None:
        """Update the enabled/visible state of each menu item with `ctx`.

        See `app_model.expressions` for details on expressions.

        Parameters
        ----------
        ctx : Mapping
            A namespace that will be used to `eval()` the `'enablement'` and
            `'when'` expressions provided for each action in the menu.
            *ALL variables used in these expressions must either be present in
            the `ctx` dict, or be builtins*.
        """
        _update_from_context(self.actions(), ctx)


def _rebuild(
    menu: QMenu | QToolBar,
    app: Application,
    menu_id: str,
    include_submenus: bool = True,
    exclude: Collection[str] | None = None,
) -> None:
    """Rebuild menu by looking up `menu` in `Application`'s menu_registry."""
    actions = menu.actions()
    for action in actions:
        menu.removeAction(action)

    _exclude = exclude or set()

    groups = list(app.menus.iter_menu_groups(menu_id))
    n_groups = len(groups)
    qapp = QApplication.instance()
    for n, group in enumerate(groups):
        for item in group:
            if isinstance(item, SubmenuItem):
                if include_submenus:
                    submenu = QModelSubmenu(item, app, parent=menu)
                    cast("QMenu", menu).addMenu(submenu)
            elif item.command.id not in _exclude:
                # use QApplication instance as parent for actions
                # because we use action singleton, and actions
                # are not related to any window.
                action = QMenuItemAction.create(item, app=app, parent=qapp)
                menu.addAction(action)
        if n < n_groups - 1:
            menu.addSeparator()


def _update_from_context(actions: Iterable[QAction], ctx: Mapping[str, object]) -> None:
    """Update the enabled/visible state of each menu item with `ctx`.

    See `app_model.expressions` for details on expressions.

    Parameters
    ----------
    actions : Iterable[QAction]
        Actions to update.
    ctx : Mapping
        A namespace that will be used to `eval()` the `'enablement'` and
        `'when'` expressions provided for each action in the menu.
        *ALL variables used in these expressions must either be present in
        the `ctx` dict, or be builtins*.
    """
    try:
        for action in actions:
            if isinstance(action, QMenuItemAction):
                action.update_from_context(ctx)
            elif isinstance(menu := action.menu(), QModelMenu):
                menu.update_from_context(ctx)
    except AttributeError as e:  # pragma: no cover
        raise AttributeError(f"This version of Qt is not supported: {e}") from e


def _find_action(actions: Iterable[QAction], object_name: str) -> QAction | None:
    return next((a for a in actions if a.objectName() == object_name), None)



import ast
from copy import deepcopy

import pytest

from app_model.expressions import Constant, Expr, Name, parse_expression, safe_eval
from app_model.expressions._expressions import _OPS, _iter_names


def test_names() -> None:
    expr = Name("n", bound=int)
    assert expr.eval({"n": 5}) == 5

    # currently, evaludating with a missing name is an error.
    with pytest.raises(NameError):
        Name("n").eval()

    assert repr(Name("n")) == "Expr.parse('n')"


def test_constants() -> None:
    assert Constant(1).eval() == 1
    assert Constant(3.14).eval() == 3.14

    assert Constant("asdf").eval() == "asdf"
    assert str(Constant("asdf")) == "'asdf'"
    assert str(Constant(r"asdf")) == "'asdf'"

    assert Constant(b"byte").eval() == b"byte"
    assert str(Constant(b"byte")) == "b'byte'"

    assert Constant(True).eval() is True
    assert Constant(False).eval() is False
    assert Constant(None).eval() is None

    assert repr(Constant(1)) == "Expr.parse('1')"

    # only {None, str, bytes, bool, int, float} allowed
    with pytest.raises(TypeError):
        Constant((1, 2))  # type: ignore


def test_bool_ops() -> None:
    n1 = Name[bool]("n1")
    true = Constant(True)
    false = Constant(False)

    assert (n1 & true).eval({"n1": True}) is True
    assert (n1 & false).eval({"n1": True}) is False
    assert (n1 & false).eval({"n1": False}) is False
    assert (n1 | true).eval({"n1": True}) is True
    assert (n1 | false).eval({"n1": True}) is True
    assert (n1 | false).eval({"n1": False}) is False

    # real constants
    assert (n1 & True).eval({"n1": True}) is True
    assert (n1 & False).eval({"n1": True}) is False
    assert (n1 & False).eval({"n1": False}) is False
    assert (n1 | True).eval({"n1": True}) is True
    assert (n1 | False).eval({"n1": True}) is True
    assert (n1 | False).eval({"n1": False}) is False

    # when working with Expr objects:
    # the binary "op" & refers to the boolean op "and"
    assert str(Constant(1) & 1) == "1 and 1"
    # note: using "and" does NOT work to combine expressions
    # (in this case, it would just return the second value "1")
    assert not isinstance(Constant(1) and 1, Expr)


def test_bin_ops() -> None:
    one = Constant(1)
    assert (one + 1).eval() == 2
    assert (one - 1).eval() == 0
    assert (one * 4).eval() == 4
    assert (one / 4).eval() == 0.25
    assert (one // 4).eval() == 0
    assert (one % 2).eval() == 1
    assert (one % 1).eval() == 0
    assert (Constant(2) ** 2).eval() == 4
    assert (one ^ 2).eval() == 3
    assert (Constant(4) & Constant(16)).eval() == 16
    assert (Constant(4) | Constant(16)).eval() == 4

    assert (Constant(16).bitand(16)).eval() == 16
    assert (Constant(16).bitor(4)).eval() == 20


def test_unary_ops() -> None:
    assert Constant(1).eval() == 1
    assert (+Constant(1)).eval() == 1
    assert (-Constant(1)).eval() == -1
    assert Constant(True).eval() is True
    assert (~Constant(True)).eval() is False


def test_comparison() -> None:
    n = Name[int]("n")
    n2 = Name[int]("n2")
    one = Constant(1)

    assert (n == n2).eval({"n": 2, "n2": 2})
    assert not (n == n2).eval({"n": 2, "n2": 1})
    assert (n != n2).eval({"n": 2, "n2": 1})
    assert not (n != n2).eval({"n": 2, "n2": 2})
    # real constant
    assert (n != 1).eval({"n": 2})
    assert not (n != 2).eval({"n": 2})

    assert (n < one).eval({"n": -1})
    assert not (n < one).eval({"n": 2})
    assert (n <= one).eval({"n": 0})
    assert (n <= one).eval({"n": 1})
    assert not (n <= one).eval({"n": 2})
    # with real constant
    assert (n < 1).eval({"n": -1})
    assert not (n < 1).eval({"n": 2})
    assert (n <= 1).eval({"n": 0})
    assert (n <= 1).eval({"n": 1})
    assert not (n <= 1).eval({"n": 2})

    assert (n > one).eval({"n": 2})
    assert not (n > one).eval({"n": 1})
    assert (n >= one).eval({"n": 2})
    assert (n >= one).eval({"n": 1})
    assert not (n >= one).eval({"n": 0})
    # real constant
    assert (n > 1).eval({"n": 2})
    assert not (n > 1).eval({"n": 1})
    assert (n >= 1).eval({"n": 2})
    assert (n >= 1).eval({"n": 1})
    assert not (n >= 1).eval({"n": 0})

    assert Expr.in_(Constant("a"), Constant("abcd")).eval() is True
    assert Constant("a").in_(Constant("abcd")).eval() is True

    assert Expr.not_in(Constant("a"), Constant("abcd")).eval() is False
    assert Constant("a").not_in(Constant("abcd")).eval() is False

    assert repr(n > n2) == "Expr.parse('n > n2')"


def test_iter_names() -> None:
    expr = "a if b in c else d > e"
    a = parse_expression(expr)
    assert a is parse_expression(a)
    b = Expr.parse(expr)  # alias
    assert sorted(_iter_names(a)) == ["a", "b", "c", "d", "e"]
    assert sorted(_iter_names(b)) == ["a", "b", "c", "d", "e"]

    with pytest.raises(RuntimeError):
        # don't directly instantiate
        Expr()


GOOD_EXPRESSIONS = [
    "a and b",
    "a == 1",
    "a @ 1",
    "2 & 4",
    "a if b == 7 else False",
    # valid constants:
    "1",
    "3.14",
    "True",
    "1 in {1, 2, 3}",
    "1 in [1, 2, 3]",
    "1 in (1, 2, 3)",
    "False",
    "None",
    "hieee",
    "b'bytes'",
    "1 < x < 2",
]

for k, v in _OPS.items():
    if issubclass(k, ast.unaryop):
        GOOD_EXPRESSIONS.append(f"{v} 1" if v == "not" else f"{v}1")
    elif v not in {"is", "is not"}:
        GOOD_EXPRESSIONS.append(f"1 {v} 2")

# these are not supported
BAD_EXPRESSIONS = [
    "a orr b",  # typo
    "a b",  # invalid syntax
    "a = b",  # Assign
    "my.attribute",  # Attribute
    "__import__(something)",  # Call
    'print("hi")',
    '{"key": "val"}',  # dicts not yet supported
    "mylist[0]",  # Index
    "mylist[0:1]",  # Slice
    'f"a"',  # JoinedStr
    "a := 1",  # NamedExpr
    r'f"{a}"',  # FormattedValue
    "[v for v in val]",  # ListComp
    "{v for v in val}",  # SetComp
    r"{k:v for k, v in val}",  # DictComp
    "(v for v in val)",  # GeneratorExp
]


@pytest.mark.parametrize("expr", GOOD_EXPRESSIONS)
def test_serdes(expr) -> None:
    assert str(parse_expression(expr)) == expr
    assert repr(parse_expression(expr))  # smoke test


@pytest.mark.parametrize("expr", BAD_EXPRESSIONS)
def test_bad_serdes(expr) -> None:
    with pytest.raises(SyntaxError):
        parse_expression(expr)


def test_deepcopy_expression() -> None:
    deepcopy(parse_expression("1"))
    deepcopy(parse_expression("1 > 2"))
    deepcopy(parse_expression("1 & 2"))
    deepcopy(parse_expression("1 or 2"))
    deepcopy(parse_expression("not 1"))
    deepcopy(parse_expression("~x"))
    deepcopy(parse_expression("2 if x else 3"))


def test_safe_eval() -> None:
    expr = "7 > x if x > 2 else 3"
    assert safe_eval(expr, {"x": 3}) is True
    assert safe_eval(expr, {"x": 10}) is False
    assert safe_eval(expr, {"x": 1}) == 3
    assert safe_eval(True) is True
    assert safe_eval(False) is False
    assert safe_eval("[1,2,3]") == [1, 2, 3]
    assert safe_eval("(1,2,3)") == (1, 2, 3)
    assert safe_eval("{1,2,3}") == {1, 2, 3}

    with pytest.raises(SyntaxError, match="Type 'Call' not supported"):
        safe_eval("func(x)")


def test_eval_kwargs() -> None:
    expr = parse_expression("a + b")
    assert expr.eval(a=1, b=2) == 3
    assert expr.eval({"a": 2}, b=2) == 4


@pytest.mark.parametrize("expr", GOOD_EXPRESSIONS)
def test_hash(expr) -> None:
    assert isinstance(hash(parse_expression(expr)), int)



import pytest
from pydantic import ValidationError

from app_model.types import Action, Icon


def test_icon_validate() -> None:
    assert Icon._validate('"fa6s.arrow_down"') == Icon(
        dark='"fa6s.arrow_down"', light='"fa6s.arrow_down"'
    )


def test_action_validation() -> None:
    with pytest.raises(ValidationError, match="'s!adf' is not a valid python_name"):
        Action(id="test", title="test", callback="s!adf")

    with pytest.raises(ValidationError):
        Action(id="test", title="test", callback=[])

    with pytest.raises(ValidationError, match=r"'x\.\<locals\>\:asdf' is not a valid"):
        Action(id="test", title="test", callback="x.<locals>:asdf")



from __future__ import annotations

import os
import sys
from collections import ChainMap
from contextlib import contextmanager
from typing import TYPE_CHECKING, Any, Callable
from weakref import finalize

from psygnal import Signal

if TYPE_CHECKING:
    from collections.abc import Iterator, MutableMapping
    from types import FrameType
    from typing import TypedDict

    class AppModelContextDict(TypedDict):
        """Global context keys offered by app-model."""

        is_linux: bool
        is_mac: bool
        is_windows: bool


_null = object()


class Context(ChainMap):
    """Evented Mapping of keys to values."""

    changed = Signal(set)  # Set[str]

    def __init__(self, *maps: MutableMapping) -> None:
        super().__init__(*maps)
        for m in maps:
            if isinstance(m, Context):
                m.changed.connect(self.changed)

    @contextmanager
    def buffered_changes(self) -> Iterator[None]:
        """Context in which to accumulated changes before emitting."""
        with self.changed.paused(lambda a, b: (a[0].union(b[0]),)):
            yield

    def __setitem__(self, k: str, v: Any) -> None:
        emit = self.get(k, _null) is not v
        super().__setitem__(k, v)
        if emit:
            self.changed.emit({k})

    def __delitem__(self, k: str) -> None:
        emit = k in self
        super().__delitem__(k)
        if emit:
            self.changed.emit({k})

    def new_child(self, m: MutableMapping | None = None) -> Context:
        """Create a new child context from this one."""
        new = super().new_child(m=m)
        self.changed.connect(new.changed)
        return new

    def __hash__(self) -> int:
        return id(self)


# note: it seems like WeakKeyDictionary would be a nice match here, but
# it appears that the object somehow isn't initialized "enough" to register
# as the same object in the WeakKeyDictionary later when queried with
# `obj in _OBJ_TO_CONTEXT` ... so instead we use id(obj)
# _OBJ_TO_CONTEXT: WeakKeyDictionary[object, Context] = WeakKeyDictionary()
_OBJ_TO_CONTEXT: dict[int, Context] = {}
_ROOT_CONTEXT: Context | None = None


def _pydantic_abort(frame: FrameType) -> bool:
    # type is being declared and pydantic is checking defaults
    # this context will never be used.
    return frame.f_code.co_name in ("__new__", "_set_default_and_type")


def create_context(
    obj: object,
    max_depth: int = 20,
    start: int = 2,
    root: Context | None = None,
    root_class: type[Context] = Context,
    frame_predicate: Callable[[FrameType], bool] = _pydantic_abort,
) -> Context:
    """Create context for any object.

    Parameters
    ----------
    obj : object
        Any object
    max_depth : int, optional
        Max frame depth to search for another object (that already has a context) off
        of which to scope this new context.  by default 20
    start : int, optional
        first frame to use in search, by default 2
    root : Optional[Context], optional
        Root context to use, by default None
    root_class : type[Context], optional
        Root class to use when creating a global root context, by default Context
        The global context is used when root is None.
    frame_predicate : Callable[[FrameType], bool], optional
        Callback that can be used to abort context creation.  Will be called on each
        frame in the stack, and if it returns True, the context will not be created.
        by default, uses pydantic-specific function to determine if a new pydantic
        BaseModel is being *declared*, (which means that the context will never be used)
        `lambda frame: frame.f_code.co_name in ("__new__", "_set_default_and_type")`

    Returns
    -------
    Optional[Context]
        Context for the object, or None if no context was found
    """
    if root is None:
        global _ROOT_CONTEXT
        if _ROOT_CONTEXT is None:
            _ROOT_CONTEXT = root_class()
        root = _ROOT_CONTEXT
    else:
        assert isinstance(root, Context), "root must be an instance of Context"

    parent = root
    if hasattr(sys, "_getframe"):  # CPython implementation detail
        frame: FrameType | None = sys._getframe(start)
        i = -1
        # traverse call stack looking for another object that has a context
        # to scope this new context off of.
        while frame and (i := i + 1) < max_depth:
            if frame_predicate(frame):
                return root  # pragma: no cover  # FIXME: should this be allowed?

            # FIXME: this might be a bit napari "magic"
            # it also assumes someone uses "self" as the first argument
            if "self" in frame.f_locals:
                _ctx = _OBJ_TO_CONTEXT.get(id(frame.f_locals["self"]))
                if _ctx is not None:
                    parent = _ctx
                    break
            frame = frame.f_back

    new_context = parent.new_child()
    obj_id = id(obj)
    _OBJ_TO_CONTEXT[obj_id] = new_context
    # remove key from dict when object is deleted
    finalize(obj, lambda: _OBJ_TO_CONTEXT.pop(obj_id, None))
    return new_context


def get_context(obj: object) -> Context | None:
    """Return context for any object, if found."""
    return _OBJ_TO_CONTEXT.get(id(obj))


def app_model_context() -> AppModelContextDict:
    """A set of useful global context keys to use."""
    return {
        "is_linux": sys.platform.startswith("linux"),
        "is_mac": sys.platform == "darwin",
        "is_windows": os.name == "nt",
    }



"""Qt objects for app_model."""

from ._qaction import QCommandAction, QCommandRuleAction, QMenuItemAction
from ._qkeybindingedit import QModelKeyBindingEdit
from ._qkeymap import (
    QKeyBindingSequence,
    qkey2modelkey,
    qkeycombo2modelkey,
    qkeysequence2modelkeybinding,
    qmods2modelmods,
)
from ._qmainwindow import QModelMainWindow
from ._qmenu import QModelMenu, QModelMenuBar, QModelSubmenu, QModelToolBar
from ._util import to_qicon

__all__ = [
    "QCommandAction",
    "QCommandRuleAction",
    "QKeyBindingSequence",
    "QMenuItemAction",
    "QModelKeyBindingEdit",
    "QModelMainWindow",
    "QModelMenu",
    "QModelMenuBar",
    "QModelSubmenu",
    "QModelToolBar",
    "qkey2modelkey",
    "qkeycombo2modelkey",
    "qkeysequence2modelkeybinding",
    "qmods2modelmods",
    "to_qicon",
]



from ._key_codes import KeyChord, KeyCode, KeyCombo, KeyMod, ScanCode
from ._keybindings import KeyBinding, SimpleKeyBinding
from ._standard_bindings import StandardKeyBinding

__all__ = [
    "KeyBinding",
    "KeyChord",
    "KeyCode",
    "KeyCombo",
    "KeyMod",
    "ScanCode",
    "SimpleKeyBinding",
    "StandardKeyBinding",
]



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, Final, cast

from psygnal import Signal

from app_model.types import MenuItem

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator

    from app_model.types import Action, DisposeCallable, MenuOrSubmenu

MenuId = str


class MenusRegistry:
    """Registry for menu and submenu items."""

    COMMAND_PALETTE_ID: Final = "_command_pallet_"
    menus_changed = Signal(set)

    def __init__(self) -> None:
        self._menu_items: dict[MenuId, dict[MenuOrSubmenu, None]] = {}

    def append_action_menus(self, action: Action) -> DisposeCallable | None:
        """Append all MenuRule items declared in `action.menus`.

        Parameters
        ----------
        action : Action
            The action containing menus to append.

        Returns
        -------
        DisposeCallable | None
            A function that can be called to unregister the menu items. If no
            menu items were registered, returns `None`.
        """
        disposers: list[Callable[[], None]] = []
        disp1 = self.append_menu_items(
            (
                rule.id,
                MenuItem(
                    command=action, when=rule.when, group=rule.group, order=rule.order
                ),
            )
            for rule in action.menus or ()
        )
        disposers.append(disp1)

        if action.palette:
            menu_item = MenuItem(command=action, when=action.enablement)
            disp = self.append_menu_items([(self.COMMAND_PALETTE_ID, menu_item)])
            disposers.append(disp)

        if not disposers:  # pragma: no cover
            return None

        def _dispose() -> None:
            for disposer in disposers:
                disposer()

        return _dispose

    def append_menu_items(
        self, items: Iterable[tuple[MenuId, MenuOrSubmenu | dict]]
    ) -> DisposeCallable:
        """Append menu items to the registry.

        Parameters
        ----------
        items : Iterable[Tuple[str, MenuOrSubmenu]]
            Items to append.

        Returns
        -------
        DisposeCallable
            A function that can be called to unregister the menu items.
        """
        changed_ids: set[str] = set()
        disposers: list[Callable[[], None]] = []
        for menu_id, item in items:
            item = cast("MenuOrSubmenu", MenuItem._validate(item))
            menu_dict = self._menu_items.setdefault(menu_id, {})
            menu_dict[item] = None
            changed_ids.add(menu_id)

            def _remove(dct: dict = menu_dict, _item: Any = item) -> None:
                dct.pop(_item, None)

            disposers.append(_remove)

        def _dispose() -> None:
            for disposer in disposers:
                disposer()
            for id_ in changed_ids:
                if not self._menu_items.get(id_):
                    del self._menu_items[id_]
            self.menus_changed.emit(changed_ids)

        if changed_ids:
            self.menus_changed.emit(changed_ids)

        return _dispose

    def __iter__(
        self,
    ) -> Iterator[tuple[MenuId, Iterable[MenuOrSubmenu]]]:
        yield from self._menu_items.items()

    def __contains__(self, id: object) -> bool:
        return id in self._menu_items

    def get_menu(self, menu_id: MenuId) -> list[MenuOrSubmenu]:
        """Return menu items for `menu_id`."""
        # using method rather than __getitem__ so that subclasses can use arguments
        return list(self._menu_items[menu_id])

    def __repr__(self) -> str:
        name = self.__class__.__name__
        return f"<{name} at {hex(id(self))} ({len(self._menu_items)} menus)>"

    def __str__(self) -> str:
        return "\n".join(self._render())

    def _render(self) -> list[str]:
        """Return registered menu items as lines of strings."""
        # this is mostly here as a debugging tool.  Can be removed or improved later.
        lines: list[str] = []

        branch = "  ├──"
        for menu in self._menu_items:
            lines.append(menu)
            for group in self.iter_menu_groups(menu):
                first = next(iter(group))
                lines.append(f"  ├───────────{first.group}───────────────")
                for child in group:
                    if isinstance(child, MenuItem):
                        lines.append(
                            f"{branch} {child.command.title} ({child.command.id})"
                        )
                    else:
                        lines.extend(
                            [
                                f"{branch} {child.submenu}",
                                "  ├──  └── ...",
                            ]
                        )
            lines.append("")
        return lines

    def iter_menu_groups(self, menu_id: MenuId) -> Iterator[list[MenuOrSubmenu]]:
        """Iterate over menu groups for `menu_id`.

        Groups are broken into sections (lists of menu or submenu items) based on
        their `group` attribute.  And each group is sorted by `order` attribute.

        Parameters
        ----------
        menu_id : str
            The menu ID to return groups for.

        Yields
        ------
        Iterator[List[MenuOrSubmenu]]
            Iterator of menu/submenu groups.
        """
        if menu_id in self:
            yield from _sort_groups(self.get_menu(menu_id))


def _sort_groups(
    items: list[MenuOrSubmenu],
    group_key: Callable = lambda x: "0000" if x == "navigation" else x or "",
    order_key: Callable = lambda x: getattr(x, "order", "") or 0,
) -> Iterator[list[MenuOrSubmenu]]:
    """Sort a list of menu items based on their .group and .order attributes."""
    groups: dict[str | None, list[MenuOrSubmenu]] = {}
    for item in items:
        groups.setdefault(item.group, []).append(item)

    for group_id in sorted(groups, key=group_key):
        yield sorted(groups[group_id], key=order_key)



import pytest

from app_model.expressions._context_keys import (
    ContextKey,
    ContextKeyInfo,
    ContextNamespace,
)


def test_context_key_info() -> None:
    key = ContextKey("default", "description", None, id="some_key")
    info = ContextKey.info()
    assert isinstance(info, list) and len(info)
    assert all(isinstance(x, ContextKeyInfo) for x in info)
    assert "some_key" in {x.key for x in info}

    assert repr(key) == "Expr.parse('some_key')"
    assert repr(key == 1) == "Expr.parse('some_key == 1')"


def _adder(x: list) -> int:
    return sum(x)


def test_context_namespace() -> None:
    class Ns(ContextNamespace):
        my_key = ContextKey[list, int](0, "description", _adder)
        optional_key = ContextKey[None, str](description="might be missing")

    assert "my_key" in Ns.__members__
    assert str(Ns.my_key) == "my_key"

    assert any(x.description == "description" for x in ContextKey.info())
    # make sure the type hints were inferred from adder
    assert Ns.my_key.__orig_class__.__args__ == (list, int)  # type: ignore

    assert isinstance(Ns.my_key, ContextKey)

    ctx: dict = {}
    ns = Ns(ctx)

    assert ns.my_key == 0
    assert ctx["my_key"] == 0
    ns.my_key = 2
    assert ctx["my_key"] == 2

    assert "optional_key" not in ctx
    assert ns.optional_key is ContextKey.MISSING
    ns.reset("optional_key")  # shouldn't raise error to reset a missing key
    # maybe the key is there though
    ctx["optional_key"] = "hi"
    assert ns.optional_key == "hi"

    ns.reset_all()
    assert ctx["my_key"] == 0
    assert "optional_key" not in ctx
    assert repr(ns) == "{'my_key': 0, 'optional_key': MISSING}"

    assert Ns.my_key.eval(ctx) == 0


def test_good_naming() -> None:
    with pytest.raises(RuntimeError):
        # you're not allowed to create a key with an id different from
        # it's attribute name
        class Ns(ContextNamespace):
            my_key = ContextKey(id="not_my_key")  # type: ignore



import re
from importlib import import_module
from typing import Any

_identifier_plus_dash = "(?:[a-zA-Z_][a-zA-Z_0-9-]+)"
_dotted_name = f"(?:(?:{_identifier_plus_dash}\\.)*{_identifier_plus_dash})"
PYTHON_NAME_PATTERN = re.compile(f"^({_dotted_name}):({_dotted_name})$")


def _validate_python_name(name: str) -> str:
    """Assert that `name` is a valid python name: e.g. `module.submodule:funcname`."""
    if name and not PYTHON_NAME_PATTERN.match(name):
        msg = (
            f"{name!r} is not a valid python_name. A python_name must "
            "be of the form '{obj.__module__}:{obj.__qualname__}' (e.g. "
            "'my_package.a_module:some_function')."
        )
        if ".<locals>." in name:  # pragma: no cover
            *_, a, b = name.split(".<locals>.")
            a = a.split(":")[-1]
            msg += (
                " Note: functions defined in local scopes are not yet supported. "
                f"Please move function {b!r} to the global scope of module {a!r}"
            )
        raise ValueError(msg)
    return name


def import_python_name(python_name: str) -> Any:
    """Import object from a fully qualified python name.

    Examples
    --------
    >>> import_python_name("my_package.a_module:some_function")
    <function some_function at 0x...>
    >>> import_python_name("pydantic:BaseModel")
    <class 'pydantic.main.BaseModel'>
    """
    _validate_python_name(python_name)  # shows the best error message
    if match := PYTHON_NAME_PATTERN.match(python_name):
        module_name, funcname = match.groups()
        mod = import_module(module_name)
        return getattr(mod, funcname)
    raise ValueError(  # pragma: no cover
        f"Could not parse python_name: {python_name!r}"
    )



import pathlib
import sys

sys.path.append(str(pathlib.Path(__file__).parent.parent))

from qtpy.QtWidgets import QApplication

from multi_file.app import MyApp

qapp = QApplication.instance() or QApplication([])
app = MyApp()
app.show()
qapp.exec()



from collections import namedtuple
from enum import Enum, auto
from typing import TYPE_CHECKING, Dict
from typing_extensions import Final

from ._key_codes import KeyCode, KeyMod

if TYPE_CHECKING:
    from .._keybinding_rule import KeyBindingRule


class StandardKeyBinding(Enum):
    AddTab = auto()
    Back = auto()
    Bold = auto()
    Cancel = auto()
    Close = auto()
    Copy = auto()
    Cut = auto()
    Delete = auto()
    DeleteCompleteLine = auto()
    DeleteEndOfLine = auto()
    DeleteEndOfWord = auto()
    DeleteStartOfWord = auto()
    Deselect = auto()
    Find = auto()
    FindNext = auto()
    FindPrevious = auto()
    Forward = auto()
    FullScreen = auto()
    HelpContents = auto()
    Italic = auto()
    MoveToEndOfDocument = auto()
    MoveToEndOfLine = auto()
    MoveToNextChar = auto()
    MoveToNextLine = auto()
    MoveToNextPage = auto()
    MoveToNextWord = auto()
    MoveToPreviousChar = auto()
    MoveToPreviousLine = auto()
    MoveToPreviousPage = auto()
    MoveToPreviousWord = auto()
    MoveToStartOfDocument = auto()
    MoveToStartOfLine = auto()
    New = auto()
    NextChild = auto()
    Open = auto()
    Paste = auto()
    Preferences = auto()
    PreviousChild = auto()
    Print = auto()
    Quit = auto()
    Redo = auto()
    Refresh = auto()
    Replace = auto()
    Save = auto()
    SaveAs = auto()
    SelectAll = auto()
    SelectEndOfDocument = auto()
    SelectEndOfLine = auto()
    SelectNextChar = auto()
    SelectNextLine = auto()
    SelectNextPage = auto()
    SelectNextWord = auto()
    SelectPreviousChar = auto()
    SelectPreviousLine = auto()
    SelectPreviousPage = auto()
    SelectPreviousWord = auto()
    SelectStartOfDocument = auto()
    SelectStartOfLine = auto()
    Underline = auto()
    Undo = auto()
    WhatsThis = auto()
    OriginalSize = auto()
    ZoomIn = auto()
    ZoomOut = auto()

    def to_keybinding_rule(self) -> "KeyBindingRule":
        """Return KeyBindingRule for this StandardKeyBinding."""
        from .._keybinding_rule import KeyBindingRule

        return KeyBindingRule(**_STANDARD_KEY_MAP[self])


_: Final[None] = None
SK = namedtuple("SK", "sk, primary, win, mac, gnome", defaults=(_, _, _, _, _))

# fmt: off
# flake8: noqa

_STANDARD_KEYS = [
    SK(StandardKeyBinding.AddTab, KeyMod.CtrlCmd | KeyCode.KeyT),
    SK(StandardKeyBinding.Back, KeyMod.Alt | KeyCode.LeftArrow, _, KeyMod.CtrlCmd | KeyCode.BracketLeft),
    SK(StandardKeyBinding.Bold, KeyMod.CtrlCmd | KeyCode.KeyB),
    SK(StandardKeyBinding.Cancel, KeyCode.Escape),
    SK(StandardKeyBinding.Close, KeyMod.CtrlCmd | KeyCode.KeyW),
    SK(StandardKeyBinding.Copy, KeyMod.CtrlCmd | KeyCode.KeyC),
    SK(StandardKeyBinding.Cut, KeyMod.CtrlCmd | KeyCode.KeyX),
    SK(StandardKeyBinding.Delete, KeyCode.Delete),
    SK(StandardKeyBinding.DeleteCompleteLine, _, _, _, KeyMod.CtrlCmd | KeyCode.KeyU),
    SK(StandardKeyBinding.DeleteEndOfLine, _, _, _, KeyMod.CtrlCmd | KeyCode.KeyK),
    SK(StandardKeyBinding.DeleteEndOfWord, _, KeyMod.CtrlCmd | KeyCode.Delete, _, KeyMod.CtrlCmd | KeyCode.Delete),
    SK(StandardKeyBinding.DeleteStartOfWord, _, KeyMod.CtrlCmd | KeyCode.Backspace, KeyMod.Alt | KeyCode.Backspace, KeyMod.CtrlCmd | KeyCode.Backspace),
    SK(StandardKeyBinding.Deselect, _, _, _, KeyMod.CtrlCmd | KeyMod.Shift | KeyCode.KeyA),
    SK(StandardKeyBinding.Find, KeyMod.CtrlCmd | KeyCode.KeyF),
    SK(StandardKeyBinding.FindNext, KeyMod.CtrlCmd | KeyCode.KeyG),
    SK(StandardKeyBinding.FindPrevious, KeyMod.CtrlCmd | KeyMod.Shift | KeyCode.KeyG),
    SK(StandardKeyBinding.Forward, _, KeyMod.Alt | KeyCode.RightArrow, KeyMod.CtrlCmd | KeyCode.BracketRight, KeyMod.Alt | KeyCode.RightArrow),
    SK(StandardKeyBinding.FullScreen, _, KeyMod.Alt | KeyCode.Enter, KeyMod.WinCtrl | KeyMod.CtrlCmd | KeyCode.KeyF, KeyMod.CtrlCmd | KeyCode.F11),
    SK(StandardKeyBinding.HelpContents, KeyCode.F1, _, KeyMod.CtrlCmd | KeyCode.Slash),
    SK(StandardKeyBinding.Italic, KeyMod.CtrlCmd | KeyCode.KeyI),
    SK(StandardKeyBinding.MoveToEndOfDocument, KeyMod.CtrlCmd | KeyCode.End, _, KeyMod.CtrlCmd | KeyCode.DownArrow),
    SK(StandardKeyBinding.MoveToEndOfLine, KeyCode.End, _, KeyMod.CtrlCmd | KeyCode.RightArrow),
    SK(StandardKeyBinding.MoveToNextChar, KeyCode.RightArrow),
    SK(StandardKeyBinding.MoveToNextLine, KeyCode.DownArrow),
    SK(StandardKeyBinding.MoveToNextPage, KeyCode.PageDown),
    SK(StandardKeyBinding.MoveToNextWord, KeyMod.CtrlCmd | KeyCode.RightArrow, _, KeyMod.Alt | KeyCode.RightArrow),
    SK(StandardKeyBinding.MoveToPreviousChar, KeyCode.LeftArrow),
    SK(StandardKeyBinding.MoveToPreviousLine, KeyCode.UpArrow),
    SK(StandardKeyBinding.MoveToPreviousPage, KeyCode.PageUp),
    SK(StandardKeyBinding.MoveToPreviousWord, KeyMod.CtrlCmd | KeyCode.LeftArrow, _, KeyMod.Alt | KeyCode.LeftArrow),
    SK(StandardKeyBinding.MoveToStartOfDocument, KeyMod.CtrlCmd | KeyCode.Home, _, KeyCode.Home),
    SK(StandardKeyBinding.MoveToStartOfLine, KeyCode.Home, _, KeyMod.CtrlCmd | KeyCode.LeftArrow),
    SK(StandardKeyBinding.New, KeyMod.CtrlCmd | KeyCode.KeyN),
    SK(StandardKeyBinding.NextChild, KeyMod.CtrlCmd | KeyCode.Tab, _, KeyMod.CtrlCmd | KeyMod.Shift | KeyCode.BracketRight),
    SK(StandardKeyBinding.Open, KeyMod.CtrlCmd | KeyCode.KeyO),
    SK(StandardKeyBinding.Paste, KeyMod.CtrlCmd | KeyCode.KeyV),
    SK(StandardKeyBinding.Preferences, KeyMod.CtrlCmd | KeyCode.Comma),
    SK(StandardKeyBinding.PreviousChild, KeyMod.CtrlCmd | KeyMod.Shift | KeyCode.Tab, _, KeyMod.CtrlCmd | KeyMod.Shift | KeyCode.BracketLeft),
    SK(StandardKeyBinding.Print, KeyMod.CtrlCmd | KeyCode.KeyP),
    SK(StandardKeyBinding.Quit, KeyMod.CtrlCmd | KeyCode.KeyQ),
    SK(StandardKeyBinding.Redo, KeyMod.CtrlCmd | KeyMod.Shift | KeyCode.KeyZ, KeyMod.CtrlCmd | KeyCode.KeyY),
    SK(StandardKeyBinding.Refresh, KeyMod.CtrlCmd | KeyCode.KeyR),
    SK(StandardKeyBinding.Replace, KeyMod.CtrlCmd | KeyCode.KeyH),
    SK(StandardKeyBinding.Save, KeyMod.CtrlCmd | KeyCode.KeyS),
    SK(StandardKeyBinding.SaveAs, KeyMod.CtrlCmd | KeyMod.Shift | KeyCode.KeyS),
    SK(StandardKeyBinding.SelectAll, KeyMod.CtrlCmd | KeyCode.KeyA),
    SK(StandardKeyBinding.SelectEndOfDocument, KeyMod.CtrlCmd | KeyMod.Shift | KeyCode.End),
    SK(StandardKeyBinding.SelectEndOfLine, KeyMod.Shift | KeyCode.End, _, KeyMod.CtrlCmd | KeyMod.Shift | KeyCode.RightArrow),
    SK(StandardKeyBinding.SelectNextChar, KeyMod.Shift | KeyCode.RightArrow),
    SK(StandardKeyBinding.SelectNextLine,  KeyMod.Shift | KeyCode.DownArrow),
    SK(StandardKeyBinding.SelectNextPage,  KeyMod.Shift | KeyCode.PageDown),
    SK(StandardKeyBinding.SelectNextWord, KeyMod.CtrlCmd | KeyMod.Shift | KeyCode.RightArrow, _, KeyMod.Alt | KeyMod.Shift | KeyCode.RightArrow),
    SK(StandardKeyBinding.SelectPreviousChar, KeyMod.Shift | KeyCode.LeftArrow),
    SK(StandardKeyBinding.SelectPreviousLine, KeyMod.Shift | KeyCode.UpArrow),
    SK(StandardKeyBinding.SelectPreviousPage, KeyMod.Shift | KeyCode.PageUp),
    SK(StandardKeyBinding.SelectPreviousWord, KeyMod.CtrlCmd | KeyMod.Shift | KeyCode.LeftArrow, _, KeyMod.Alt | KeyMod.Shift | KeyCode.LeftArrow),
    SK(StandardKeyBinding.SelectStartOfDocument, KeyMod.CtrlCmd | KeyMod.Shift | KeyCode.Home),
    SK(StandardKeyBinding.SelectStartOfLine, KeyMod.Shift | KeyCode.Home, _, KeyMod.CtrlCmd | KeyMod.Shift | KeyCode.LeftArrow),
    SK(StandardKeyBinding.Underline, KeyMod.CtrlCmd | KeyCode.KeyU),
    SK(StandardKeyBinding.Undo, KeyMod.CtrlCmd | KeyCode.KeyZ),
    SK(StandardKeyBinding.WhatsThis, KeyMod.Shift | KeyCode.F1),
    SK(StandardKeyBinding.OriginalSize, KeyMod.CtrlCmd | KeyCode.Digit0),
    SK(StandardKeyBinding.ZoomIn, KeyMod.CtrlCmd | KeyCode.Equal),
    SK(StandardKeyBinding.ZoomOut, KeyMod.CtrlCmd | KeyCode.Minus),
]

# fmt: on
_STANDARD_KEY_MAP: Dict[StandardKeyBinding, Dict[str, str]] = {
    nt.sk: {"primary": nt.primary, "win": nt.win, "mac": nt.mac, "linux": nt.gnome}
    for nt in _STANDARD_KEYS
}



from __future__ import annotations

import contextlib
from typing import TYPE_CHECKING, ClassVar
from weakref import WeakValueDictionary

from qtpy.QtGui import QKeySequence

from app_model import Application
from app_model.expressions import Expr
from app_model.types import ToggleRule

from ._qkeymap import QKeyBindingSequence
from ._util import to_qicon

if TYPE_CHECKING:
    from collections.abc import Mapping

    from PyQt6.QtGui import QAction
    from qtpy.QtCore import QObject
    from typing_extensions import Self

    from app_model.types import CommandRule, MenuItem
else:
    from qtpy.QtWidgets import QAction


class QCommandAction(QAction):
    """Base QAction for a command id. Can execute the command.

    Parameters
    ----------
    command_id : str
        Command ID.
    app : Application | str
        Application instance or name of application instance.
    parent : QObject | None
        Optional parent widget, by default None
    """

    def __init__(
        self,
        command_id: str,
        app: Application | str,
        parent: QObject | None = None,
    ) -> None:
        super().__init__(parent)
        self._app = Application.get_or_create(app) if isinstance(app, str) else app
        self._command_id = command_id
        self.setObjectName(command_id)
        self._keybinding_tooltip = ""
        if kb := self._app.keybindings.get_keybinding(command_id):
            self.setShortcut(QKeyBindingSequence(kb.keybinding))
            self._keybinding_tooltip = f"({kb.keybinding.to_text()})"
        self.triggered.connect(self._on_triggered)

    def _update_keybinding(self) -> None:
        shortcut = self.shortcut()
        if kb := self._app.keybindings.get_keybinding(self._command_id):
            self.setShortcut(QKeyBindingSequence(kb.keybinding))
            self._keybinding_tooltip = f"({kb.keybinding.to_text()})"
        elif not shortcut.isEmpty():
            self.setShortcut(QKeySequence())
            self._keybinding_tooltip = ""

    def _on_triggered(self, checked: bool) -> None:
        # execute_command returns a Future, for the sake of eventually being
        # asynchronous without breaking the API.  For now, we call result()
        # to raise any exceptions.
        self._app.commands.execute_command(self._command_id).result()


class QCommandRuleAction(QCommandAction):
    """QAction for a CommandRule.

    Parameters
    ----------
    command_rule : CommandRule
        `CommandRule` instance to create an action for.
    app : Application | str
        Application instance or name of application instance.
    parent : QObject | None
        Optional parent widget, by default None
    use_short_title : bool
        If True, use the `short_title` of the command rule, if it exists.
    """

    def __init__(
        self,
        command_rule: CommandRule,
        app: Application | str,
        parent: QObject | None = None,
        *,
        use_short_title: bool = False,
    ) -> None:
        super().__init__(command_rule.id, app, parent)
        self._cmd_rule = command_rule
        self._tooltip = command_rule.tooltip or ""
        if use_short_title and command_rule.short_title:
            self.setText(command_rule.short_title)  # pragma: no cover
        else:
            self.setText(command_rule.title)
        if command_rule.icon:
            self.setIcon(to_qicon(command_rule.icon))
        self.setIconVisibleInMenu(command_rule.icon_visible_in_menu)
        if command_rule.status_tip:
            self.setStatusTip(command_rule.status_tip)
        if command_rule.toggled is not None:
            self.setCheckable(True)
            self._refresh()
        tooltip_with_keybinding = f"{self._tooltip} {self._keybinding_tooltip}".rstrip()
        self.setToolTip(tooltip_with_keybinding)

    def setText(self, text: str | None) -> None:
        super().setText(text)
        self._tooltip = self._tooltip or text or ""

    def _update_keybinding(self) -> None:
        super()._update_keybinding()
        tooltip_with_keybinding = f"{self._tooltip} {self._keybinding_tooltip}".rstrip()
        self.setToolTip(tooltip_with_keybinding)

    def update_from_context(self, ctx: Mapping[str, object]) -> None:
        """Update the enabled state of this menu item from `ctx`."""
        self.setEnabled(expr.eval(ctx) if (expr := self._cmd_rule.enablement) else True)
        if expr2 := self._cmd_rule.toggled:
            if isinstance(expr2, Expr) or (
                isinstance(expr2, ToggleRule) and (expr2 := expr2.condition)
            ):
                self.setChecked(expr2.eval(ctx))

    def _refresh(self) -> None:
        if isinstance(self._cmd_rule.toggled, ToggleRule):
            if get_current := self._cmd_rule.toggled.get_current:
                _current = self._app.injection_store.inject(
                    get_current, on_unresolved_required_args="ignore"
                )
                self.setChecked(_current())


class QMenuItemAction(QCommandRuleAction):
    """QAction for a MenuItem.

    Mostly the same as a `CommandRuleAction`, but aware of the `menu_item.when` clause
    to toggle visibility.

    Parameters
    ----------
    menu_item : MenuItem
        `MenuItem` instance to create an action for.
    app : Application | str
        Application instance or name of application instance.
    parent : QObject | None
        Optional parent widget, by default None
    """

    _cache: ClassVar[WeakValueDictionary[tuple[int, int], Self]] = WeakValueDictionary()

    def __init__(
        self,
        menu_item: MenuItem,
        app: Application | str,
        parent: QObject | None = None,
    ) -> None:
        super().__init__(menu_item.command, app, parent)
        self._menu_item = menu_item
        with contextlib.suppress(NameError):
            self.update_from_context(self._app.context)

    @staticmethod
    def _cache_key(app: Application, menu_item: MenuItem) -> tuple[int, int]:
        return (id(app), hash(menu_item))

    @classmethod
    def create(
        cls,
        menu_item: MenuItem,
        app: Application | str,
        parent: QObject | None = None,
    ) -> Self:
        """Create a new QMenuItemAction for the given menu item.

        Prefer this method over `__init__` to ensure that the cache is used,
        so that:

        ```python
        a1 = QMenuItemAction.create(action, full_app)
        a2 = QMenuItemAction.create(action, full_app)
        a1 is a2  # True
        ```
        """
        app = Application.get_or_create(app) if isinstance(app, str) else app
        cache_key = QMenuItemAction._cache_key(app, menu_item)
        if cache_key in cls._cache:
            res = cls._cache[cache_key]
            res._update_keybinding()
            res.setParent(parent)
            return res

        cls._cache[cache_key] = obj = cls(menu_item, app, parent)
        return obj

    def update_from_context(self, ctx: Mapping[str, object]) -> None:
        """Update the enabled/visible state of this menu item from `ctx`."""
        super().update_from_context(ctx)
        self.setVisible(expr.eval(ctx) if (expr := self._menu_item.when) else True)

    def __repr__(self) -> str:
        name = self.__class__.__name__
        return f"{name}({self._menu_item!r}, app={self._app.name!r})"



import runpy
from pathlib import Path

import pytest
from qtpy.QtWidgets import QApplication

DEMO = Path(__file__).parent.parent.parent / "demo"


@pytest.mark.parametrize("fname", ["qapplication.py", "model_app.py", "multi_file"])
def test_qapp(qapp, fname, monkeypatch) -> None:
    monkeypatch.setattr(QApplication, "exec", lambda *a, **k: None)
    runpy.run_path(str(DEMO / fname), run_name="__main__")



import sys

from qtpy.QtWidgets import QApplication

from app_model.backends.qt import QModelKeyBindingEdit

app = QApplication(sys.argv)
w = QModelKeyBindingEdit()
w.editingFinished.connect(lambda: print(w.keyBinding()))
w.show()
sys.exit(app.exec())



from app_model.types import Action, KeyBindingRule, KeyCode, KeyMod, MenuRule

from . import functions
from .constants import CommandId, MenuId

ACTIONS: list[Action] = [
    Action(
        id=CommandId.OPEN,
        title="Open",
        icon="fa6-solid:folder-open",
        callback=functions.open_file,
        menus=[MenuRule(id=MenuId.FILE)],
        keybindings=[KeyBindingRule(primary=KeyMod.CtrlCmd | KeyCode.KeyO)],
    ),
    Action(
        id=CommandId.CLOSE,
        title="Close",
        icon="fa-solid:window-close",
        callback=functions.close,
        menus=[MenuRule(id=MenuId.FILE)],
        keybindings=[KeyBindingRule(primary=KeyMod.CtrlCmd | KeyCode.KeyW)],
    ),
    Action(
        id=CommandId.UNDO,
        title="Undo",
        icon="fa-solid:undo",
        callback=functions.undo,
        menus=[MenuRule(id=MenuId.EDIT, group="1_undo_redo")],
        keybindings=[KeyBindingRule(primary=KeyMod.CtrlCmd | KeyCode.KeyZ)],
    ),
    Action(
        id=CommandId.REDO,
        title="Redo",
        icon="fa6-solid:rotate-right",
        callback=functions.redo,
        menus=[MenuRule(id=MenuId.EDIT, group="1_undo_redo")],
        keybindings=[
            KeyBindingRule(primary=KeyMod.CtrlCmd | KeyMod.Shift | KeyCode.KeyZ)
        ],
    ),
    Action(
        id=CommandId.CUT,
        title="Cut",
        icon="fa-solid:cut",
        callback=functions.cut,
        menus=[MenuRule(id=MenuId.EDIT, group="3_copypaste")],
        keybindings=[KeyBindingRule(primary=KeyMod.CtrlCmd | KeyCode.KeyX)],
    ),
    Action(
        id=CommandId.COPY,
        title="Copy",
        icon="fa6-solid:copy",
        callback=functions.copy,
        menus=[MenuRule(id=MenuId.EDIT, group="3_copypaste")],
        keybindings=[KeyBindingRule(primary=KeyMod.CtrlCmd | KeyCode.KeyC)],
    ),
    Action(
        id=CommandId.PASTE,
        title="Paste",
        icon="fa6-solid:paste",
        callback=functions.paste,
        menus=[MenuRule(id=MenuId.EDIT, group="3_copypaste")],
        keybindings=[KeyBindingRule(primary=KeyMod.CtrlCmd | KeyCode.KeyV)],
    ),
]



import pytest

from app_model import Application
from app_model.registries import register_action
from app_model.types import Action, KeyBinding

PRIMARY_KEY = "ctrl+a"
OS_KEY = "ctrl+b"
MENUID = "some.menu.id"
KWARGS = [
    {},
    {"enablement": "x == 1"},
    {"menus": [MENUID]},  # test that we can pass menus as a single string too
    {"enablement": "3 >= 1", "menus": [{"id": MENUID}]},
    {"keybindings": [{"primary": PRIMARY_KEY}]},
    {
        "keybindings": [
            {"primary": PRIMARY_KEY, "mac": OS_KEY, "win": OS_KEY, "linux": OS_KEY}
        ]
    },
    {"keybindings": [{"primary": "ctrl+a"}], "menus": [{"id": MENUID}]},
    {"palette": False},
]


@pytest.mark.parametrize("kwargs", KWARGS)
@pytest.mark.parametrize("mode", ["str", "decorator", "action"])
def test_register_action_decorator(
    kwargs: dict, simple_app: Application, mode: str
) -> None:
    # make sure mocks are working
    app = simple_app
    assert not list(app.commands)
    assert not list(app.keybindings)
    assert not list(app.menus)

    cmd_id = "cmd.id"
    kwargs["title"] = "Test title"

    # register the action
    if mode == "decorator":

        @register_action(app=app, id_or_action=cmd_id, **kwargs)
        def f1() -> str:
            return "hi"

        assert f1() == "hi"  # decorator returns the function

    else:

        def f2() -> str:
            return "hi"

        if mode == "str":
            register_action(app=app, id_or_action=cmd_id, callback=f2, **kwargs)

        elif mode == "action":
            action = Action(id=cmd_id, callback=f2, **kwargs)
            app.register_action(action)

    # make sure the command is registered
    assert cmd_id in app.commands
    assert list(app.commands)
    # make sure an event was emitted signaling the command was registered
    app.commands_changed.assert_called_once_with(cmd_id)  # type: ignore

    # make sure we can call the command, and that we can inject dependencies.
    assert app.commands.execute_command(cmd_id).result() == "hi"

    # make sure menus are registered if specified
    menus = kwargs.get("menus", [])
    if menus := kwargs.get("menus"):
        for entry in menus:
            id_ = entry if isinstance(entry, str) else entry["id"]
            assert id_ in app.menus
            app.menus_changed.assert_any_call({id_})
    else:
        menus = list(app.menus)
        if kwargs.get("palette") is not False:
            assert app.menus.COMMAND_PALETTE_ID in app.menus
            assert len(menus) == 1
        else:
            assert not list(app.menus)

    # make sure keybindings are registered if specified
    if keybindings := kwargs.get("keybindings"):
        for entry in keybindings:
            key = PRIMARY_KEY if len(entry) == 1 else OS_KEY  # see KWARGS[5]
            key = KeyBinding.from_str(key)
            assert any(i.keybinding == key for i in app.keybindings)
            app.keybindings_changed.assert_called()  # type: ignore
    else:
        assert not list(app.keybindings)

    # check that calling the dispose function removes everything.
    app.dispose()
    assert not list(app.commands)
    assert not list(app.keybindings)
    assert not list(app.menus)


def test_errors(simple_app: Application) -> None:
    with pytest.raises(ValueError, match="'title' is required"):
        simple_app.register_action("cmd_id")  # type: ignore
    with pytest.raises(TypeError, match="must be a string or an Action"):
        simple_app.register_action(None)  # type: ignore


def test_register_multiple_actions(simple_app: Application) -> None:
    actions: list[Action] = [
        Action(id="cmd_id1", title="title1", callback=lambda: None),
        Action(id="cmd_id2", title="title2", callback=lambda: None),
    ]
    dispose = simple_app.register_actions(actions)
    assert len(simple_app.commands) == 2
    dispose()
    assert not list(simple_app.commands)



from qtpy.QtWidgets import QApplication, QFileDialog


def open_file() -> None:
    name, _ = QFileDialog.getOpenFileName()
    print("Open file:", name)


def close() -> None:
    if win := QApplication.activeWindow():
        win.close()
    print("close")


def undo() -> None:
    print("undo")


def redo() -> None:
    print("redo")


def cut() -> None:
    print("cut")


def copy() -> None:
    print("copy")


def paste() -> None:
    print("paste")



from __future__ import annotations

import sys
from typing import TYPE_CHECKING, cast

import pytest
from qtpy.QtCore import Qt
from qtpy.QtWidgets import QAction, QMainWindow

from app_model.backends.qt import QModelMenu, QModelToolBar

if TYPE_CHECKING:
    from pytestqt.plugin import QtBot

    from conftest import FullApp

SEP = ""
LINUX = sys.platform.startswith("linux")


@pytest.mark.parametrize("MenuCls", [QModelMenu, QModelToolBar])
def test_menu(
    MenuCls: type[QModelMenu] | type[QModelToolBar], qtbot: QtBot, full_app: FullApp
) -> None:
    app = full_app

    menu = MenuCls(app.Menus.EDIT, app, title="just-for-coverage")
    qtbot.addWidget(menu)

    # The "" are separators, according to our group settings in full_app
    menu_texts = [a.text() for a in menu.actions()]
    assert menu_texts == ["AtTop", SEP, "Undo", "Redo", SEP, "Copy", "Paste"]

    # check that triggering the actions calls the associated commands
    for cmd in (app.Commands.UNDO, app.Commands.REDO):
        action = cast("QAction", menu.findAction(cmd))
        with qtbot.waitSignal(action.triggered):
            action.trigger()
            getattr(app.mocks, cmd).assert_called_once()

    redo_action = cast("QAction", menu.findAction(app.Commands.REDO))

    assert redo_action.isVisible()
    assert redo_action.isEnabled()

    # change that visibility and enablement follows the context
    menu.update_from_context({"allow_undo_redo": True, "something_to_undo": False})
    assert redo_action.isVisible()
    assert redo_action.isEnabled()

    menu.update_from_context({"allow_undo_redo": False, "something_to_undo": False})
    assert redo_action.isVisible()
    assert not redo_action.isEnabled()

    menu.update_from_context({"allow_undo_redo": False, "something_to_undo": True})
    assert not redo_action.isVisible()
    assert not redo_action.isEnabled()

    menu.update_from_context({"allow_undo_redo": True, "something_to_undo": False})
    assert redo_action.isVisible()
    assert redo_action.isEnabled()

    # useful error when we forget a required name
    with pytest.raises(NameError, match="Names required to eval this expression"):
        menu.update_from_context({})

    menu._disconnect()


def test_submenu(qtbot: QtBot, full_app: FullApp) -> None:
    app = full_app

    menu = QModelMenu(app.Menus.FILE, app)
    qtbot.addWidget(menu)

    menu_texts = [a.text() for a in menu.actions()]
    assert menu_texts == ["Open From...", "Open..."]

    submenu = menu.findChild(QModelMenu, app.Menus.FILE_OPEN_FROM)
    submenu_action = submenu.findAction(app.Commands.OPEN_FROM_B)

    assert submenu_action
    assert isinstance(submenu, QModelMenu)
    submenu.setVisible(True)
    assert submenu.isVisible()
    assert submenu.isEnabled()
    menu.aboutToShow.emit()  # for test coverage

    # "not something_open" is the when clause
    # "friday" is the enablement clause

    menu.update_from_context({"something_open": False, "friday": True, "sat": True})
    assert submenu.isVisible()
    assert submenu.isEnabled()
    assert submenu_action.isVisible()
    assert submenu_action.isEnabled()

    menu.update_from_context({"something_open": False, "friday": False, "sat": True})
    assert submenu.isVisible()
    assert not submenu.isEnabled()
    assert submenu_action.isVisible()
    assert submenu_action.isEnabled()

    menu.update_from_context({"something_open": True, "friday": False, "sat": True})
    assert not submenu.isEnabled()
    assert submenu_action.isEnabled()

    menu.update_from_context({"something_open": True, "friday": True, "sat": True})
    assert submenu.isEnabled()
    assert submenu_action.isEnabled()

    menu.update_from_context({"something_open": True, "friday": True, "sat": False})
    assert submenu.isEnabled()
    assert not submenu_action.isEnabled()


@pytest.mark.filterwarnings("ignore:QPixmapCache.find:")
@pytest.mark.skipif(LINUX, reason="Linux keytest not working")
def test_shortcuts(qtbot: QtBot, full_app: FullApp) -> None:
    app = full_app

    win = QMainWindow()
    menu = QModelMenu(app.Menus.EDIT, app=app, title="Edit", parent=win)
    win.menuBar().addMenu(menu)
    qtbot.addWidget(win)
    qtbot.addWidget(menu)

    with qtbot.waitExposed(win):
        win.show()

    copy_action = menu.findAction(app.Commands.COPY)

    with qtbot.waitSignal(copy_action.triggered, timeout=2000):
        qtbot.keyClicks(win, "C", Qt.KeyboardModifier.ControlModifier)

    paste_action = menu.findAction(app.Commands.PASTE)
    with qtbot.waitSignal(paste_action.triggered, timeout=1000):
        qtbot.keyClicks(win, "V", Qt.KeyboardModifier.ControlModifier)


def test_toggled_menu_item(qtbot: QtBot, full_app: FullApp) -> None:
    app = full_app
    menu = QModelMenu(app.Menus.HELP, app)
    qtbot.addWidget(menu)

    menu.update_from_context({"thing_toggled": True})
    action = menu.findAction(app.Commands.TOGGLE_THING)
    assert action.isChecked()

    menu.update_from_context({"thing_toggled": False})
    assert not action.isChecked()


@pytest.mark.parametrize("MenuCls", [QModelMenu, QModelToolBar])
def test_menu_events(
    MenuCls: type[QModelMenu] | type[QModelToolBar], qtbot: QtBot, full_app: FullApp
) -> None:
    app = full_app
    menu = MenuCls(app.Menus.EDIT, app)

    qtbot.addWidget(menu)

    # The "" are separators, according to our group settings in full_app
    menu_texts = [a.text() for a in menu.actions()]
    assert menu_texts == ["AtTop", SEP, "Undo", "Redo", SEP, "Copy", "Paste"]

    # simulate something changing the edit menu... normally this would be
    # triggered by a dispose() call, but that's a bit hard to do currently with the
    # test app fixture.
    copy_item = next(
        x for x in full_app.menus._menu_items["edit"] if x.command.title == "Copy"
    )
    full_app.menus._menu_items["edit"].pop(copy_item)
    full_app.menus.menus_changed.emit(app.Menus.EDIT)

    menu_texts = [a.text() for a in menu.actions()]
    assert menu_texts == ["AtTop", SEP, "Undo", "Redo", SEP, "Paste"]



# mypy: disable-error-code="var-annotated"
import pytest

from app_model.registries import KeyBindingsRegistry, MenusRegistry
from app_model.registries._keybindings_reg import _RegisteredKeyBinding
from app_model.types import (
    Action,
    KeyBinding,
    KeyBindingRule,
    KeyBindingSource,
    KeyCode,
    KeyMod,
    MenuItem,
)


def _noop() -> None:
    pass


def test_menus_registry() -> None:
    reg = MenusRegistry()
    reg.append_menu_items([("file", {"command": {"id": "file.new", "title": "File"}})])
    reg.append_menu_items([("file.sub", {"submenu": "Sub", "title": "SubTitle"})])

    assert isinstance(reg.get_menu("file")[0], MenuItem)
    assert "(2 menus)" in repr(reg)
    assert "File" in str(reg)
    assert "Sub" in str(reg)  # ok to change


def test_keybindings_registry() -> None:
    reg = KeyBindingsRegistry()
    assert "(0 bindings)" in repr(reg)


def test_register_keybinding_rule_filter_type() -> None:
    """Check `_filter_keybinding` type checking when setting."""
    reg = KeyBindingsRegistry()
    with pytest.raises(TypeError, match="'filter_keybinding' must be a callable"):
        reg.filter_keybinding = "string"  # type: ignore


def _filter_fun(kb: KeyBinding) -> str:
    if kb.part0.is_modifier_key():
        return "modifier only sequences not allowed"
    return ""


def test_register_keybinding_rule_filter_get() -> None:
    """Check `_filter_keybinding` getter."""
    reg = KeyBindingsRegistry()
    reg.filter_keybinding = _filter_fun
    assert callable(reg.filter_keybinding)


def test_register_keybinding_rule_filter() -> None:
    """Check `filter_keybinding` in `register_keybinding_rule`."""
    reg = KeyBindingsRegistry()
    reg.filter_keybinding = _filter_fun

    # Valid keybinding
    kb = KeyBindingRule(primary=KeyMod.CtrlCmd | KeyCode.KeyO)
    reg.register_keybinding_rule("test", kb)
    # Invalid keybinding
    kb = KeyBindingRule(primary=KeyMod.Alt)
    with pytest.raises(ValueError, match=r"Alt\+: modifier only"):
        reg.register_keybinding_rule("test", kb)


@pytest.mark.parametrize(
    "kb, msg",
    [
        (
            [
                {"primary": KeyMod.CtrlCmd | KeyCode.KeyA},
                {"primary": KeyMod.Shift | KeyCode.KeyC},
            ],
            "",
        ),
        (
            [{"primary": KeyMod.Alt}, {"primary": KeyMod.Shift}],
            r"Alt\+: modifier only sequences not allowed\nShift\+: modifier",
        ),
    ],
)
def test_register_action_keybindings_filter(kb: list[dict], msg: str) -> None:
    """Check `filter_keybinding` in `register_action_keybindings`."""
    reg = KeyBindingsRegistry()
    reg.filter_keybinding = _filter_fun

    action = Action(
        id="cmd_id1",
        title="title1",
        callback=_noop,
        keybindings=kb,
    )
    if msg:
        with pytest.raises(ValueError, match=msg):
            reg.register_action_keybindings(action)
    else:
        reg.register_action_keybindings(action)


@pytest.mark.parametrize(
    "kb1, kb2, kb3",
    [
        (
            [
                {
                    "primary": KeyMod.CtrlCmd | KeyCode.KeyA,
                    "when": "active",
                    "weight": 10,
                },
            ],
            [
                {
                    "primary": KeyMod.CtrlCmd | KeyCode.KeyA,
                },
            ],
            [
                {"primary": KeyMod.CtrlCmd | KeyCode.KeyA, "weight": 5},
            ],
        ),
    ],
)
def test_register_action_keybindings_prioritization(
    kb1: str, kb2: str, kb3: str
) -> None:
    """Check `get_context_prioritized_keybinding`."""
    reg = KeyBindingsRegistry()

    action1 = Action(
        id="cmd_id1",
        title="title1",
        callback=_noop,
        keybindings=kb1,
    )
    reg.register_action_keybindings(action1)

    action2 = Action(
        id="cmd_id2",
        title="title2",
        callback=_noop,
        keybindings=kb2,
    )
    reg.register_action_keybindings(action2)

    action3 = Action(
        id="cmd_id3",
        title="title3",
        callback=_noop,
        keybindings=kb3,
    )
    reg.register_action_keybindings(action3)

    keybinding = reg.get_context_prioritized_keybinding(
        kb1[0]["primary"], {"active": False}
    )
    assert keybinding.command_id == "cmd_id3"

    keybinding = reg.get_context_prioritized_keybinding(
        kb1[0]["primary"], {"active": True}
    )
    assert keybinding.command_id == "cmd_id1"

    keybinding = reg.get_context_prioritized_keybinding(
        KeyMod.Shift | kb1[0]["primary"], {"active": True}
    )
    assert keybinding is None


@pytest.mark.parametrize(
    "kb1, kb2, gt, lt, eq",
    [
        (
            {
                "primary": KeyMod.CtrlCmd | KeyCode.KeyA,
                "command_id": "command_id1",
                "weight": 0,
                "when": None,
                "source": KeyBindingSource.USER,
            },
            {
                "primary": KeyMod.CtrlCmd | KeyCode.KeyA,
                "command_id": "command_id2",
                "weight": 0,
                "when": None,
                "source": KeyBindingSource.APP,
            },
            True,
            False,
            False,
        ),
        (
            {
                "primary": KeyMod.CtrlCmd | KeyCode.KeyA,
                "command_id": "command_id1",
                "weight": 0,
                "when": None,
                "source": KeyBindingSource.USER,
            },
            {
                "primary": KeyMod.CtrlCmd | KeyCode.KeyA,
                "command_id": "command_id2",
                "weight": 10,
                "when": None,
                "source": KeyBindingSource.APP,
            },
            True,
            False,
            False,
        ),
        (
            {
                "primary": KeyMod.CtrlCmd | KeyCode.KeyA,
                "command_id": "command_id1",
                "weight": 0,
                "when": None,
                "source": KeyBindingSource.USER,
            },
            {
                "primary": KeyMod.CtrlCmd | KeyCode.KeyA,
                "command_id": "command_id2",
                "weight": 10,
                "when": None,
                "source": KeyBindingSource.USER,
            },
            False,
            True,
            False,
        ),
        (
            {
                "primary": KeyMod.CtrlCmd | KeyCode.KeyA,
                "command_id": "command_id1",
                "weight": 10,
                "when": None,
                "source": KeyBindingSource.USER,
            },
            {
                "primary": KeyMod.CtrlCmd | KeyCode.KeyA,
                "command_id": "command_id2",
                "weight": 10,
                "when": None,
                "source": KeyBindingSource.USER,
            },
            False,
            False,
            True,
        ),
    ],
)
def test_registered_keybinding_comparison(
    kb1: dict, kb2: dict, gt: bool, lt: bool, eq: bool
) -> None:
    rkb1 = _RegisteredKeyBinding(
        keybinding=kb1["primary"],
        command_id=kb1["command_id"],
        weight=kb1["weight"],
        when=kb1["when"],
        source=kb1["source"],
    )
    rkb2 = _RegisteredKeyBinding(
        keybinding=kb2["primary"],
        command_id=kb2["command_id"],
        weight=kb2["weight"],
        when=kb2["when"],
        source=kb2["source"],
    )
    assert (rkb1 > rkb2) == gt
    assert (rkb1 < rkb2) == lt
    assert (rkb1 == rkb2) == eq
